# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
VCS interface (application and documents)
"""


import sys
import os
import importlib

class VCSError(Exception):
    pass

# dict holding references to repo modules
# initially strings with module names,
# after first call to is_available the values
# point to the modules themselves.
_vcs_modules = {
    'git': "vcs.gitrepo"
}

def is_available(tool):
    """Returns True if the requested VCS software is available on the system.

    Supported vcs are:
    - 'git'
    - so far this is the only one

    Any NNNrepo.py module has to implement a function vcs_available()
    """
    if tool not in _vcs_modules.keys():
        raise VCSError('Invalid argument for VCS software: {}\nSupported:\n- {}'.format(
            tool,
            "\n- ".join(_vcs_modules.keys())
        ))
    mod = _vcs_modules[tool]
    if type(mod) is str:
        mod = _vcs_modules[tool] = importlib.import_module(mod)
    return mod.Repo.vcs_available()


def app_is_git_controlled():
    """Return True if Frescobaldi is running from Git.

    This is done by checking for the presence of the .git/ directory.
    The function is very cheap, the directory test is only performed on the
    first call.

    """
    global _app_is_git_controlled
    try:
        return _app_is_git_controlled
    except NameError:
        _app_is_git_controlled = True
        if os.path.isdir(os.path.join(sys.path[0], '..', '.git')):
            if is_available('git'):
                from . import apprepo
                global app_repo
                app_repo = apprepo.Repo()
            else:
                from PyQt6.QtWidgets import QMessageBox
                QMessageBox.warning(None,
                                    _("Git not found"),
                                    _("Frescobaldi is run from within a Git "
                                      "repository, but Git does not appear "
                                      "to be working. Git support will be "
                                      "disabled. If you have Git installed, "
                                      "you can specify its location in the "
                                      "Preferences dialog."))
                _app_is_git_controlled = False
        else:
            _app_is_git_controlled = False
        return _app_is_git_controlled

# conditionally create app_repo object
app_is_git_controlled()

if app_is_git_controlled():
    # debugging tests, to be removed finally
    #from . import test
    pass

def app_active_branch_window_title():
    """Return the active branch, suitable as window title.

    If the app is not git-controlled, the empty string is returned.

    """
    if app_is_git_controlled():
        git_branch = app_repo.active_branch()
        return '({branch} [{remote}])'.format(
                branch=git_branch,
                remote=app_repo.tracked_remote_label(git_branch))
    return ''




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Settings stuff and handling for different LilyPond versions.
"""


import glob
import codecs
import os
import pathlib
import platform
import re
import shutil

from PyQt6.QtCore import QEventLoop, QSettings, QStandardPaths, QTimer
from PyQt6.QtWidgets import QProgressDialog

import app
import cachedproperty
import job
import job.queue
import util
import qutil

if platform.system() == "Darwin":
    import macos

_infos = None   # this can hold a list of configured LilyPondInfo instances

LILYPOND_AUTOINSTALL_DIR = (
    os.path.join(QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppLocalDataLocation),
                 "lilypond-binaries")
)
# According to the Qt documentation, the location may need to be created.
pathlib.Path(LILYPOND_AUTOINSTALL_DIR).mkdir(parents=True, exist_ok=True)

def infos():
    """Returns all configured LilyPondInfo for the different used LilyPond versions."""
    global _infos
    if _infos is None:
        s = QSettings()
        _infos = []
        for i in range(s.beginReadArray("lilypondinfo")):
            s.setArrayIndex(i)
            info = LilyPondInfo.read(s)
            if info:
                _infos.append(info)
        s.endArray()
        if not _infos:
            info = default()
            if info.abscommand():
                _infos.append(info)
        app.aboutToQuit.connect(saveinfos)
    return _infos


def setinfos(infos):
    """Sets the info's to the given list of LilyPondInfo instances."""
    global _infos
    _infos = infos


def saveinfos():
    """Saves the info's."""
    s = QSettings()
    s.beginWriteArray("lilypondinfo")
    s.remove("")    # previously saved infos may no longer be valid
    i = 0           # see below for why we don't use enumerate() here
    for info in infos():
        s.setArrayIndex(i)
        try:
            info.write(s)
            # QSettings expects consecutive array indices, so we don't want
            # to increment this for unavailable versions, which leaves a gap
            i += 1
        except FileNotFoundError:
            # this LilyPond version is no longer available
            _infos.remove(info)
    s.endArray()


def default():
    """Returns a default LilyPondInfo instance with the default LilyPond command.

    On Windows, the default command is "lilypond.exe",
    on other platforms simply "lilypond".

    """
    lilypond = "lilypond.exe" if platform.system() == "Windows" else "lilypond"
    return LilyPondInfo(lilypond)


def preferred():
    """Returns the preferred (user set default) LilyPondInfo instance."""
    infos_ = infos()
    if not infos_:
        return default()
    elif len(infos_) == 1:
        return infos_[0]
    s = QSettings()
    s.beginGroup("lilypond_settings")
    # find default version
    defaultCommand = "lilypond.exe" if platform.system() == "Windows" else "lilypond"
    userDefault = s.value("default", defaultCommand, str)
    if userDefault != defaultCommand:
        for info in infos_:
            if info.command == userDefault:
                return info
    for info in infos_:
        if info.command == defaultCommand:
            return info
    return infos_[0]


def suitable(version):
    """Return a LilyPondInfo with a suitable version if found.

    Otherwise the most recent LilyPond version is returned.

    """
    infos_ = [i for i in infos() if i.autoVersionIncluded]
    if infos_:
        infos_.sort(key=lambda i: i.version())
        for i in infos_:
            if i.version() >= version:
                return i
        return i # return the latest anyway
    return preferred()


class CachedProperty(cachedproperty.CachedProperty):
    def wait(self, msg=None, timeout=0):
        """Returns the value for the property, waiting for it to be computed.

        If this lasts longer than 2 seconds, a progress dialog is displayed.

        """
        if self.get() is None:
            self.start()
            if self.get() is None:
                if msg is None:
                    msg = _("Running LilyPond, this can take some time...")
                qutil.waitForSignal(self.computed, msg, timeout)
        return self.get()

    __call__ = wait


class LilyPondInfo:
    ly_tool_names = (
        'lilypond-book',
        'convert-ly',
        'musicxml2ly',
        'midi2ly',
        'abc2ly',
    )

    def __init__(self, command):
        self._command = command
        self.autoVersionIncluded = True
        self.name = "LilyPond"
        self._lytools = {}

    @property
    def command(self):
        return self._command

    @CachedProperty.cachedproperty
    def abscommand(self):
        """The absolute path of the command."""
        if platform.system() == "Windows":
            # on Windows, newer versions of LilyPond don't add themselves to the
            # PATH, so add a probable path here
            path = glob.glob(os.path.join(
                os.environ.get('ProgramFiles', 'C:\\Program Files'),
                'LilyPond*', 'usr', 'bin'))
            if not path:
                # LilyPond may be built for a different architecture
                bits, linkage = platform.architecture()
                if bits.startswith('64'):
                    # 64-bit Frescobaldi with 32-bit LilyPond
                    path = glob.glob(os.path.join(
                        os.environ.get('ProgramFiles(x86)', 'C:\\Program Files (x86)'),
                        'LilyPond*', 'usr', 'bin'))
                elif bits.startswith('32'):
                    # 32-bit Frescobaldi with 64-bit LilyPond
                    path = glob.glob(os.path.join(
                        os.environ.get('ProgramW6432', 'C:\\Program Files'),
                        'LilyPond*', 'usr', 'bin'))
        elif platform.system() == "Darwin":
            # also on macOS, LilyPond is not automatically added to the PATH
            path = [
                os.path.join('/Applications', 'LilyPond.app', 'Contents', 'Resources', 'bin'),
                os.path.join('/opt', 'local', 'bin'),
                os.path.join('/opt', 'lilypond', 'bin'),
                # Default path for Homebrew on MacOS arm64.
                '/opt/homebrew/bin',
                # Default path for Homebrew on MacOS x86_64.
                '/usr/local/bin',
            ]
            if macos.inside_app_bundle():
                # By default, the PATH environment variable for app bundles
                # does not contain the path which is required for lilypond
                # and other helper programs like gs.
                # Add the lilypond path to PATH to find the helper programs
                # for Homebrew, Macports and other local installations.
                exe = util.findexe(self.command, path)
                if exe:
                    bindir = os.path.dirname(exe)
                    os.environ['PATH'] = bindir + ':' + os.environ['PATH']
        else:
            path = None
        return util.findexe(self.command, path) or False

    #NOTE/TODO:
    # This has only been tested for downloaded and self-compiled releases on Linux so far
    @CachedProperty.cachedproperty(depends=abscommand)
    def libdir(self):
        exe = self.abscommand()
        if exe:
            parent = os.path.dirname(os.path.dirname(exe))
            return os.path.join(parent, 'lib')
        else:
            return False

    @CachedProperty.cachedproperty(depends=abscommand)
    def displaycommand(self):
        """The path to the command in a format pretty to display.

        This removes the 'out/bin/lilypond' part of custom build LilyPond
        versions, and on macOS it removes the
        '/Contents/Resources/bin/lilypond' part.

        Finally it replaces the users home directory with '~'.

        The empty string is returned if LilyPond is not installed on the users'
        system.
        """
        command = self.abscommand()
        if command:
            outstrip='out/bin/lilypond'
            if command.endswith(outstrip):
                command=command[:-len(outstrip)]
            macstrip='/Contents/Resources/bin/lilypond'
            if platform.system() == "Darwin" and command.endswith('.app' + macstrip):
                command=command[:-len(macstrip)]
            return util.homify(command)
        else:
            return self.command

    @CachedProperty.cachedproperty(depends=abscommand)
    def versionString(self):
        if not self.abscommand():
            return ""

        j = job.Job([self.abscommand(), '--version'])

        @j.done.connect
        def done():
            success = j.success
            if success:
                output = ' '.join([line[0] for line in j.history()])
                m = re.search(r"\d+\.\d+(.\d+)?", output)
                self.versionString = m.group() if m else ""
            else:
                self.versionString = ""

        app.job_queue().add_job(j, 'generic')

    @CachedProperty.cachedproperty(depends=versionString)
    def version(self):
        if self.versionString():
            return tuple(map(int, self.versionString().split('.')))
        return ()

    @CachedProperty.cachedproperty(depends=abscommand)
    def bindir(self):
        """Returns the directory the LilyPond command is in."""
        if self.abscommand():
            return os.path.dirname(self.abscommand())
        return False

    @CachedProperty.cachedproperty(depends=bindir)
    def prefix(self):
        """Returns the prefix LilyPond was installed to."""
        if self.bindir():
            return os.path.dirname(self.bindir())
        return False

    @CachedProperty.cachedproperty(depends=(prefix, versionString))
    def datadir(self):
        """Returns the datadir of this LilyPond instance.

        Most times this is something like "/usr/share/lilypond/2.13.3/"
        If this method returns False, the datadir could not be determined.

        """
        if not self.abscommand():
            return False

        # First ask LilyPond itself.
        j = job.Job([self.abscommand(), '-e',
            "(display (ly:get-option 'datadir)) (newline) (exit)"])
        @j.done.connect
        def done():
            success = j.success
            if success:
                output = [line[0] for line in j.history()]
                d = output[1].strip('\n')
                if os.path.isabs(d) and os.path.isdir(d):
                    self.datadir = d
                    return

            # Then find out via the prefix.
            if self.prefix():
                dirs = ['current']
                if self.versionString():
                    dirs.append(self.versionString())
                for suffix in dirs:
                    d = os.path.join(self.prefix(), 'share', 'lilypond', suffix)
                    if os.path.isdir(d):
                        self.datadir = d
                        return
            self.datadir = False
        app.job_queue().add_job(j, 'generic')

    def toolcommand(self, original_command, use_ly_tool=True):
        """Return a list containing the commandline to run a tool, e.g. convert-ly.

        On Unix (including macOS), the list has one element: the full path to the tool.

        On Windows, the list has two elements: the LilyPond-provided Python
        interpreter and the tool path.

        If use_ly_tool is True, this takes into account the command the user
        might have configured for the tool, using ly_tool() to get that command
        name first.

        """
        if use_ly_tool:
            command = self.ly_tool(original_command)
        else:
            command = original_command

        bindir = self.bindir()
        if bindir:
            toolpath = os.path.join(self.bindir(), command)
        else:
            toolpath = command

        # on Windows the tool command is not directly executable, but
        # must be started using the LilyPond-provided Python interpreter
        if platform.system() == "Windows":
            if not os.access(toolpath, os.R_OK) and not toolpath.endswith('.py'):
                toolpath += '.py'
            command = [self.python(), toolpath]
        else:
            command = [toolpath]
        return command

    @property
    def isAutoManaged(self):
        """Return true if the LilyPond installation was auto-installed by Frescobaldi."""
        command_path = pathlib.Path(self.command)
        autoinstall_path = pathlib.Path(LILYPOND_AUTOINSTALL_DIR)
        #return command_path.is_relative_to(autoinstall_path) # Python 3.9+
        return autoinstall_path in command_path.parents

    def forget(self):
        """If this is an auto-managed installation, then wipe it from the file system."""
        if self.isAutoManaged:
            try:
                shutil.rmtree(self.prefix())
            except FileNotFoundError:
                pass    # this installation has already been deleted

    @CachedProperty.cachedproperty(depends=versionString)
    def prettyName(self):
        """Return a pretty-printable name for this LilyPond instance."""
        # Don't show the internal path for auto-managed installations
        if self.isAutoManaged:
            return "{name} {version}".format(name=self.name,
                                             version=self.versionString())
        else:
            return "{name} {version} ({command})".format(
                name = self.name,
                version = self.versionString(),
                command = self.displaycommand())

    @CachedProperty.cachedproperty(depends=version)
    def lilydoc_url(self):
        """Return the URL of this LilyPond version's online documentation."""
        # see also appinfo.lilydoc_stable and appinfo.lilydoc_development
        version = self.version()
        if len(version) >= 2:
            return "http://lilypond.org/doc/v{major}.{minor}".format(
                major=version[0], minor=version[1])
        else:
            return "http://lilypond.org/doc"

    def ly_tool(self, name):
        """Get the configured command for the ly tool (e.g. midi2ly).

        By default this is just the name itself, relative to the directory
        the LilyPond executable is in.

        """
        return self._lytools.get(name, name)

    def set_ly_tool(self, name, value):
        """Set the command for the named ly tool (e.g. midi2ly).

        By default this is just the name itself, relative to the directory
        the LilyPond executable is in.

        """
        self._lytools[name] = value

    @classmethod
    def read(cls, settings):
        """Returns a new LilyPondInfo instance, filled from a QSettings instance.

        May return None, if the command is not existing.

        """
        cmd = settings.value("command", "", str)
        if cmd:
            info = cls(cmd)
            if info.abscommand.wait():
                info.autoVersionIncluded = settings.value("auto", True, bool)
                info.name = settings.value("name", "LilyPond", str)
                for name in cls.ly_tool_names:
                    info.set_ly_tool(name, settings.value(name, name, str))
                if int(os.path.getmtime(info.abscommand())) == int(settings.value("mtime", 0, float)):
                    info.versionString = settings.value("version", "", str)
                    datadir = settings.value("datadir", "", str)
                    if datadir and os.path.isdir(datadir):
                        info.datadir = datadir
                return info

    def write(self, settings):
        """Writes ourselves to a QSettings instance. We should be valid."""
        settings.setValue("command", self.command)
        settings.setValue("version", self.versionString())
        settings.setValue("datadir", self.datadir() or "")
        if self.abscommand():
            settings.setValue("mtime", int(os.path.getmtime(self.abscommand())))
        settings.setValue("auto", self.autoVersionIncluded)
        settings.setValue("name", self.name)
        for name in self.ly_tool_names:
            value = self._lytools.get(name, name)
            if name == value:
                settings.remove(name)
            else:
                settings.setValue(name, value)

    def python(self):
        """Returns the path to the LilyPond-provided Python interpreter.

        This is only used on Windows, where tools like convert-ly can't be
        run directly.

        """
        if self.bindir():
            for python in ('python-windows.exe', 'pythonw.exe', 'python.exe'):
                interpreter = os.path.join(self.bindir(), python)
                if os.access(interpreter, os.X_OK):
                    return interpreter
        return 'pythonw.exe'



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Information about the Frescobaldi application.
"""


# these variables are also used by the distutils setup
name = "frescobaldi"
# https://packaging.python.org/en/latest/discussions/versioning/
version = "4.0.6"
extension_api = "0.9.0"
description = "LilyPond Music Editor"
long_description = \
    "Frescobaldi is an advanced text editor to edit LilyPond sheet music files. " \
    "Features include an integrated PDF preview and a powerful Score Wizard."
maintainer = "Wilbert Berendsen"
maintainer_email = "frescobaldi@googlegroups.com"
domain = "frescobaldi.org"
url = f"http://www.{domain}/"
issues_url = "https://github.com/frescobaldi/frescobaldi/issues/new"
license = "GPL"

# this one is used everywhere in the application
appname = "Frescobaldi"

desktop_file_name = "org.frescobaldi.Frescobaldi"

# required versions of important dependencies
required_python_version = (3, 10)
required_python_ly_version = (0, 9, 4)
required_qt_version = (6, 6)

# LilyPond doc URLs to be used in lilydoc/manager.py
# see also lilypondinfo.LilyPondInfo.lilydoc_url()
lilydoc_stable = "http://lilypond.org/doc/v2.24"
lilydoc_development = "http://lilypond.org/doc/v2.25"



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Expand variables like $DATE, $LILYPOND_VERSION etc. in snippets.
"""


import time

import appinfo
import lilypondinfo


ANCHOR, CURSOR, SELECTION = constants = 1, 2, 3 # just some constants

class Expander:
    """Expands variables.

    The methods return text or other events (currently simply integer constants).

    Each variable is documented in the user guide.
    """
    def __init__(self, cursor):
        self.cursor = cursor

    def DATE(self):
        return time.strftime('%Y-%m-%d')

    def LILYPOND_VERSION(self):
        return lilypondinfo.preferred().versionString()

    def FRESCOBALDI_VERSION(self):
        return appinfo.version

    def URL(self):
        return self.cursor.document().url().toString()

    def FILE_NAME(self):
        return self.cursor.document().url().toLocalFile()

    def DOCUMENT_NAME(self):
        return self.cursor.document().documentName()

    def CURSOR(self):
        return CURSOR

    def ANCHOR(self):
        return ANCHOR

    def SELECTION(self):
        return SELECTION if self.cursor.hasSelection() else CURSOR



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Import non-lilypond file types.
"""


import os
import importlib
import shutil

from PyQt6.QtCore import QUrl
from PyQt6.QtGui import QAction, QTextCursor
from PyQt6.QtWidgets import QFileDialog, QMessageBox

import app
import icons
import actioncollection
import actioncollectionmanager
import plugin
import util
import job.dialog


class FileImport(plugin.MainWindowPlugin):

    def __init__(self, mainwindow):
        self._import_job = None
        self._import_dialog = None
        self._import_file = ''
        self._mxml_dialog = None
        self._midi_dlg = None
        self._abc_dlg = None
        self.targets = {
            '.xml': ('.musicxml', self._mxml_dialog),
            '.musicxml': ('.musicxml', self._mxml_dialog),
            '.mxl': ('.musicxml', self._mxml_dialog),
            '.midi': ('.midi', self._midi_dlg),
            '.mid': ('.midi', self._midi_dlg),
            '.abc': ('.abc', self._abc_dlg)
        }

        ac = self.actionCollection = Actions()
        actioncollectionmanager.manager(mainwindow).addActionCollection(ac)
        ac.import_any.triggered.connect(self.import_any)
        ac.import_musicxml.triggered.connect(self.import_musicxml)
        ac.import_midi.triggered.connect(self.import_midi)
        ac.import_abc.triggered.connect(self.import_abc)

    def configure_import(self, input_file):
        """Open the configuration dialog corresponding to the file type."""
        self._import_file = input_file

        # Retrieve module and dialog corresponding to import file type
        ext = os.path.splitext(self._import_file)[1]
        module_name, dlg = self.targets[ext]
        if not dlg:
            # create dialog suitable for import file type
            module = importlib.import_module(module_name, 'file_import')
            dlg = module.Dialog(self.mainwindow())
        dlg.set_input(self._import_file)
        self._import_dialog = dlg

    def do_import(self, filetypes, caption, multiple=False):
        """Control structure for file import.
        Can either import "any" file(s) or one file of a specific file type.
        """
        self._import_files = self.get_import_file(filetypes, caption, multiple)
        for filename in self._import_files:
            # TODO: Update the file dialog to not allow "All files" anymore
            # (probably requires the use of a QSortFilterProxyModel).
            # Then is_importable and the following conditional can be
            # removed.
            if self.is_importable(filename):
                self.configure_import(filename)
                self.run_import()
            else:
                QMessageBox.critical(None, _("Error"),
                    _("The file {filename} could not be converted: "
                      "wrong file type.").format(filename=filename))

    def get_import_file(self, filetypes, caption, multiple=False):
        """Open a File Open dialog for the requested filetype(s),
        return a list with one or multiple existing file names
        or an empty list in case the user has canceled the dialog."""
        caption = app.caption(caption)
        directory = (os.path.dirname(
            self.mainwindow().currentDocument().url().toLocalFile())
            or app.basedir())

        if multiple:
            result = QFileDialog.getOpenFileNames(
                self.mainwindow(), caption, directory, filetypes)[0]
            return result if result else []
        else:
            result = QFileDialog.getOpenFileName(
                self.mainwindow(), caption, directory, filetypes)[0]
            return [result] if result else []

    def import_any(self):
        """Reads the file type and determines which import to use."""
        filetypes = ';;'.join((
            '{} (*.xml *.musicxml *.mxl *.midi *.mid *.abc)'.format(_("All importable formats")),
            '{} (*.xml *.musicxml *.mxl)'.format(_("MusicXML Files")),
            '{} (*.midi *.mid)'.format(_("Midi Files")),
            '{} (*.abc)'.format(_("ABC Files")),
            '{} (*)'.format(_("All Files")),
        ))
        caption = app.caption(_("dialog title", "Import"))
        self.do_import(filetypes, caption, multiple=True)

    def import_musicxml(self):
        """Opens a MusicXML file. Converts it to ly by using musicxml2ly."""
        filetypes = '{} (*.xml *.musicxml *.mxl);;{} (*)'.format(
            _("MusicXML Files"), _("All Files"))
        caption = _("dialog title", "Import a MusicXML file")
        self.do_import(filetypes, caption)

    def import_midi(self):
        """Opens an midi file. Converts it to ly by using midi2ly."""
        filetypes = '{} (*.midi *.mid);;{} (*)'.format(
            _("Midi Files"), _("All Files"))
        caption = app.caption(_("dialog title", "Import a midi file"))
        self.do_import(filetypes, caption)

    def import_abc(self):
        """Opens an abc file. Converts it to ly by using abc2ly."""
        filetypes = '{} (*.abc);;{} (*)'.format(
            _("ABC Files"), _("All Files"))
        caption = app.caption(_("dialog title", "Import an abc file"))
        self.do_import(filetypes, caption)

    def import_done(self):
        j = self._import_job
        conf_dlg = self._import_dialog
        conf_dlg.saveSettings()
        lyfile = os.path.splitext(self._import_file)[0] + ".ly"
        while (os.path.exists(lyfile)
            or app.findDocument(QUrl.fromLocalFile(lyfile))):
            lyfile = util.next_file(lyfile)
        shutil.move(j.output_file(), lyfile)

        doc = app.openUrl(QUrl.fromLocalFile(lyfile))
        doc.setModified(True)
        self.mainwindow().setCurrentDocument(doc)

        self.post_import(conf_dlg.get_post_settings(), doc)
        self.mainwindow().saveDocument(doc)

    def is_importable(self, filename):
        """Check if the file is importable."""
        ext = os.path.splitext(filename)[1]
        return ext in ['.xml', '.musicxml', '.mxl', '.midi', '.mid', '.abc']

    def post_import(self, settings, doc):
        """Adaptations of the source after running musicxml2ly

        Present settings:
        Reformat source
        Remove superfluous durations
        Remove duration scaling
        Engrave directly
        """
        cursor = QTextCursor(doc)
        if settings[0]:
            import reformat
            reformat.reformat(cursor)
        if settings[1]:
            cursor.select(QTextCursor.SelectionType.Document)
            from rhythm import rhythm
            rhythm.rhythm_implicit_per_line(cursor)
        if settings[2]:
            cursor.select(QTextCursor.SelectionType.Document)
            from rhythm import rhythm
            rhythm.rhythm_remove_fraction_scaling(cursor)
        if settings[3]:
            import engrave
            engrave.engraver(self.mainwindow()).engrave('preview', doc, False)

    def run_import(self):
        """Generic execution of all import dialogs."""
        conf_dlg = self._import_dialog
        if conf_dlg.exec():
            self._import_job = j = conf_dlg.job()
            dlg = job.dialog.Dialog(self.mainwindow(), auto_accept=True)
            dlg.accepted.connect(self.import_done)
            dlg.run(j)


class Actions(actioncollection.ActionCollection):
    name = "file_import"
    def createActions(self, parent):
        self.import_any = QAction(parent)
        self.import_musicxml = QAction(parent)
        self.import_midi = QAction(parent)
        self.import_abc = QAction(parent)

        self.import_any.setIcon(icons.get("document-import"))

    def translateUI(self):
        self.import_any.setText(_("Import..."))
        self.import_any.setToolTip(_("Generic import for all LilyPond tools."))
        self.import_musicxml.setText(_("Import MusicXML..."))
        self.import_musicxml.setToolTip(_("Import a MusicXML file using musicxml2ly."))
        self.import_midi.setText(_("Import Midi..."))
        self.import_midi.setToolTip(_("Import a Midi file using midi2ly."))
        self.import_abc.setText(_("Import abc..."))
        self.import_abc.setToolTip(_("Import an abc file using abc2ly."))



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Setup the application language.

Also contains a function to get language preferences from the operation
system.

"""

import locale

from PyQt6.QtCore import QLocale, QSettings, QTimer
from PyQt6.QtWidgets import QMessageBox

import app

from . import install, available, UnknownLanguageError

__all__ = ['preferred', 'current', 'default']


_currentlanguage = None


def preferred():
    """Return a list of language codes from the operating system preferences.

    Language- and country codes will always be separated with an underscore '_'.

    """
    try:
        langs = QLocale().uiLanguages()
    except AttributeError:
        # QLocale.uiLanguages is not in Qt 4.7 (only Qt4.8+)
        langs = []
    else:
        # in some systems, language/country codes have '-' and not '_'
        langs = [l.replace('-', '_') for l in langs]
    if not langs:
        try:
            langs.append(locale.getdefaultlocale()[0])
        except ValueError:
            pass
    return langs

def default():
    """Return the first preferred system default UI language that is available in Frescobaldi.

    May return "C", if none of the system preferred languages is available
    in Frescobaldi. Note that this is effectively equivalent to a return of "en",
    though the latter means that English is actually requested by the user,
    not a fallback.

    """
    av_langs = available()
    av_langs.append("en")
    for lang in preferred():
        if lang in av_langs or lang.split('_')[0] in av_langs:
            return lang
    return "C"

def current():
    """Returns the currently active UI language code.

    A name is always returned, which can be "C" or "en", meaning no
    translation is desired.

    """
    return QSettings().value("language", "", str) or default()

def _setup():
    """Set application language according to settings."""
    global _currentlanguage
    new_language = current()
    try:
        install(new_language)
    except UnknownLanguageError:
        import language_names
        lang_name = language_names.languageName(new_language, "en")
        msg = """\
Frescobaldi was configured to display its interface in {language}, but the \
translation is not available. Frescobaldi will continue in English. This probably \
means you installed Frescobaldi from source and are missing the MO catalogs; \
please generate them using: \
<code>tox -e mo-generate</code>""".format(language=lang_name)
        QMessageBox.critical(None, app.caption("Error"), msg)
        new_language = "C"
        install(new_language)
    if _currentlanguage is not None and new_language != _currentlanguage:
            QTimer.singleShot(0, app.languageChanged)
    _currentlanguage = new_language


@app.oninit
def _start_up():
    """Initialize GUI translations. Called op app startup."""
    _setup()
    from . import qtranslator
    qtranslator.initialize()


app.settingsChanged.connect(_setup)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The Documents menu.
"""


from PyQt6.QtGui import QAction, QActionGroup, QIcon
from PyQt6.QtWidgets import QMenu

import app
import icons
import plugin
import engrave
import documenticon
import util


class DocumentMenu(QMenu):
    def __init__(self, mainwindow):
        super().__init__(mainwindow)
        self.aboutToShow.connect(self.populate)
        app.translateUI(self)
        self.setToolTipsVisible(True)

    def translateUI(self):
        self.setTitle(_('menu title', '&Documents'))

    def populate(self):
        self.clear()
        mainwindow = self.parentWidget()
        for a in DocumentActionGroup.instance(mainwindow).actions():
            self.addAction(a)


class DocumentActionGroup(plugin.MainWindowPlugin, QActionGroup):
    """Maintains a list of actions to set the current document.

    The actions are added to the View->Documents menu in the order
    of the tabbar. The actions also get accelerators that are kept
    during the lifetime of a document.

    """
    def __init__(self, mainwindow):
        QActionGroup.__init__(self, mainwindow)
        self._acts = {}
        self._accels = {}
        self.setExclusive(True)
        for d in app.documents:
            self.addDocument(d)
        app.documentCreated.connect(self.addDocument)
        app.documentClosed.connect(self.removeDocument)
        app.documentUrlChanged.connect(self.setDocumentStatus)
        app.documentModificationChanged.connect(self.setDocumentStatus)
        app.jobStarted.connect(self.setDocumentStatus)
        app.jobFinished.connect(self.setDocumentStatus)
        mainwindow.currentDocumentChanged.connect(self.setCurrentDocument)
        engrave.engraver(mainwindow).stickyChanged.connect(self.setDocumentStatus)
        self.triggered.connect(self.slotTriggered)

    def actions(self):
        return [self._acts[doc] for doc in self.mainwindow().documents()]

    def addDocument(self, doc):
        a = QAction(self)
        a.setCheckable(True)
        if doc is self.mainwindow().currentDocument():
            a.setChecked(True)
        self._acts[doc] = a
        self.setDocumentStatus(doc)

    def removeDocument(self, doc):
        self._acts[doc].deleteLater()
        del self._acts[doc]
        del self._accels[doc]

    def setCurrentDocument(self, doc):
        self._acts[doc].setChecked(True)

    def setDocumentStatus(self, doc):
        # create accels
        accels = [self._accels[d] for d in self._accels if d is not doc]
        name = doc.documentName().replace('&', '&&')
        for index, char in enumerate(name):
            if char.isalnum() and char.lower() not in accels:
                name = name[:index] + '&' + name[index:]
                self._accels[doc] = char.lower()
                break
        else:
            self._accels[doc] = ''
        # add [sticky] mark if necessary
        if doc == engrave.engraver(self.mainwindow()).stickyDocument():
            # L10N: 'always engraved': the document is marked as 'Always Engrave' in the LilyPond menu
            name += " " + _("[always engraved]")
        self._acts[doc].setText(name)
        self._acts[doc].setToolTip(util.path(doc.url()))
        icon = documenticon.icon(doc, self.mainwindow())
        if icon.name() == "text-plain":
            icon = QIcon()
        self._acts[doc].setIcon(icon)

    def slotTriggered(self, action):
        for doc, act in self._acts.items():
            if act == action:
                self.mainwindow().setCurrentDocument(doc)
                break



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Base class for viewers' context menus.
"""

import os

from PyQt6.QtCore import QObject, QUrl
from PyQt6.QtGui import QAction, QActionGroup
from PyQt6.QtWidgets import QApplication, QMenu

import app
import icons


class AbstractViewerContextMenu(QObject):
    """Base class for viewers' context menus.
    It provides the template method pattern to generate the menu.
    Subclasses can override individual parts of the menu this way."""

    # Has to be set in subclasses, used to ask for extension menu
    name = ""

    def __init__(self, panel):
        self._panel = panel
        self._actionCollection = panel.actionCollection
        self._menu = QMenu(self._panel)

    def addSeparator(self):
        """Add a separator to the menu."""
        self._menu.addSeparator()

    def addCopyImageAction(self):
        """Add action to copy image if available"""
        if self._panel.widget().view.rubberband().hasSelection():
            self._menu.addAction(self._actionCollection.viewer_copy_image)

    def addEditInPlaceAction(self, cursor, position):
        """Add action to edit snippet if on a textedit link"""
        a = self._menu.addAction(icons.get("document-edit"), _("Edit in Place"))
        @a.triggered.connect
        def edit():
            import editinplace
            editinplace.edit(self._panel.widget(), cursor, position)

    def addLinkAction(self, link):
        """Add action if on an arbitrary link"""
        m = self._menu
        a = m.addAction(icons.get("window-new"), _("Open Link in &New Window"))
        @a.triggered.connect
        def open_in_browser():
            import helpers
            helpers.openUrl(QUrl(link.url))

        a = m.addAction(icons.get("edit-copy"), _("Copy &Link"))
        @a.triggered.connect
        def copy_link():
            QApplication.clipboard().setText(link.url)

    def addCursorLinkActions(self):
        """Add actions if on a textedit or arbitrary link"""
        if self._cursor:
            self.addEditInPlaceAction(self._cursor, self._position)
        elif self._link and self._link.url:
            self.addLinkAction(self._link)

    def addExtensionMenu(self):
        """Add a submenu with all actions exposed by extensions"""
        self._menu.addMenu(app.extensions().menu(self.name))

    def addShowActions(self):
        """Adds a submenu giving access to the (other)
        opened viewer documents"""
        mds = self._actionCollection.viewer_document_select
        docs = mds.viewdocs()
        multi_docs = len(docs) > 1
        if self._panel.widget().currentViewdoc():
            current_doc_filename = self._panel.widget().currentViewdoc().filename()

        m = self._menu
        sm = QMenu(m)
        sm.setTitle(_("Show..."))
        sm.setEnabled(multi_docs)
        ag = QActionGroup(m)
        ag.triggered.connect(self._panel.slotShowViewdoc)

        for d in docs:
            action = QAction(sm)
            action.setText(os.path.basename(d.filename()))
            action._document_filename = d.filename()
            # TODO: Tooltips aren't shown by Qt (it seems)
            action.setToolTip(d.filename())
            action.setCheckable(True)
            action.setChecked(d.filename() == current_doc_filename)

            ag.addAction(action)
            sm.addAction(action)

        m.addSeparator()
        m.addMenu(sm)

    def addOpenCloseActions(self):
        """Add actions to close documents.
        This is not implemented in the base class"""
        m = self._menu
        ac = self._actionCollection
        m.addAction(ac.viewer_open)
        docs = self._actionCollection.viewer_document_select.viewdocs()
        if docs:
            sm = QMenu(m)
            sm.setTitle(_("Close..."))
            m.addMenu(sm)
            sm.addAction(ac.viewer_close)
            multi_docs = len(docs) > 1
            ac.viewer_close_other.setEnabled(multi_docs)
            ac.viewer_close_all.setEnabled(multi_docs)
            sm.addAction(ac.viewer_close_other)
            sm.addAction(ac.viewer_close_all)

    def addReloadAction(self):
        """Add action to reload document."""
        current_document = self._panel.widget().currentViewdoc()
        if current_document:
            m = self._menu
            ac = self._actionCollection
            m.addAction(ac.viewer_reload)

    def addZoomActions(self):
        """Add actions to zoom the viewer"""
        m = self._menu
        m.addSeparator()
        sm = QMenu(m)
        sm.setTitle(_("Zoom"))
        m.addMenu(sm)
        ac = self._actionCollection
        sm.addAction(ac.viewer_fit_width)
        sm.addAction(ac.viewer_fit_height)
        sm.addAction(ac.viewer_fit_both)
        sm.addSeparator()
        sm.addAction(ac.viewer_zoom_in)
        sm.addAction(ac.viewer_zoom_out)
        sm.addAction(ac.viewer_zoom_original)

    def addSynchronizeAction(self):
        """Add an action telling the viewer to
        always try syncing with the input editor."""
        m = self._menu
        ac = self._actionCollection
        m.addAction(ac.viewer_sync_cursor)

    def addShowToolbarAction(self):
        """Add action to toggle the visibility of
        the viewer's toolbar"""
        m = self._menu
        ac = self._actionCollection
        m.addAction(ac.viewer_show_toolbar)

    def addHelpAction(self):
        """Add help menu item"""
        m = self._menu
        ac = self._actionCollection
        m.addSeparator()
        m.addAction(ac.viewer_help)

    def show(self, position, link, cursor, methods = None):
        """Build the panel's context menu dynamically.
        Implements the template method pattern to allow
        subclasses to override each step.
        If methods passes a list of methods these will be
        called instead to construct the menu, making it
        possible to change order or add methods not
        available in the base class."""

        self._position = position
        self._link = link
        self._cursor = cursor

        self._menu.clear()

        if not methods:
            # Actions affecting the current link(selection)
            self.addCopyImageAction()
            self.addCursorLinkActions()
            # Actions affecting the currently opened documents
            self.addShowActions()
            self.addOpenCloseActions()
            self.addReloadAction()
            # Actions affecting the viewer's state
            self.addZoomActions()
            self.addSynchronizeAction()
            self.addShowToolbarAction()
            self.addExtensionMenu()
        else:
            for m in methods:
                m()
        # The help action is added always
        self.addHelpAction()

        # show it!
        if self._menu.actions():
            self._menu.exec(position)
        # TODO: Can this really be removed?
        # it had to be commented out because in the documents submenu
        # there popped up issues with deleted objects (the menu was already
        # deleted when the signal was emitted)
#        self._menu.deleteLater()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The Preferences Dialog.
"""



from PyQt6.QtCore import QMargins, QSettings, QSize, Qt, pyqtSignal
from PyQt6.QtGui import QKeySequence
from PyQt6.QtWidgets import (
    QDialog, QDialogButtonBox, QGroupBox, QHBoxLayout, QListWidget,
    QListWidgetItem, QScrollArea, QStackedWidget, QVBoxLayout, QWidget)

import app
import qutil
import userguide
import icons
import widgets

_prefsindex = 0 # global setting for selected prefs page but not saved on exit

def pageorder():
    """Yields the page item classes in order."""
    yield General
    yield LilyPond
    yield MusicViewers
    yield Midi
    yield Editor
    yield Tools
    yield Paths
    yield Documentation
    yield Shortcuts
    yield FontsColors
    yield Helpers
    yield Extensions


class CancelClosingPreferences(Exception):
    pass


class PreferencesDialog(QDialog):

    def __init__(self, mainwindow):
        super().__init__(mainwindow)
        self.setWindowModality(Qt.WindowModality.WindowModal)
        if mainwindow:
            self.addAction(mainwindow.actionCollection.help_whatsthis)
        layout = QVBoxLayout()
        layout.setSpacing(10)
        self.setLayout(layout)

        # listview to the left, stacked widget to the right
        top = QHBoxLayout()
        layout.addLayout(top)

        self.pagelist = QListWidget(self)
        self.stack = QStackedWidget(self)
        top.addWidget(self.pagelist, 0)
        top.addWidget(self.stack, 2)

        layout.addWidget(widgets.Separator(self))

        b = self.buttons = QDialogButtonBox(self)
        b.setStandardButtons(
            QDialogButtonBox.StandardButton.Ok
            | QDialogButtonBox.StandardButton.Cancel
            | QDialogButtonBox.StandardButton.Apply
            | QDialogButtonBox.StandardButton.Reset
            | QDialogButtonBox.StandardButton.Help)
        layout.addWidget(b)
        b.accepted.connect(self.maybeAccept)
        b.rejected.connect(self.reject)
        # saveSettings() may raise CancelClosingPreferences. This is primarily
        # for use when it makes sense for the "Ok" button, but may occur in "Apply",
        # so we need to catch it.
        b.button(QDialogButtonBox.StandardButton.Apply).clicked.connect(self.trySaveSettings)
        b.button(QDialogButtonBox.StandardButton.Reset).clicked.connect(self.loadSettings)
        b.button(QDialogButtonBox.StandardButton.Help).clicked.connect(self.showHelp)
        b.button(QDialogButtonBox.StandardButton.Help).setShortcut(QKeySequence.StandardKey.HelpContents)
        b.button(QDialogButtonBox.StandardButton.Apply).setEnabled(False)

        # fill the pagelist
        self.pagelist.setIconSize(QSize(32, 32))
        self.pagelist.setSpacing(2)
        for item in pageorder():
            self.pagelist.addItem(item())
        self.pagelist.currentItemChanged.connect(self.slotCurrentItemChanged)

        app.translateUI(self, 100)
        # read our size and selected page
        self.pagelist.setCurrentRow(_prefsindex)
        qutil.saveDialogSize(self, "preferences/dialog/size", self.sizeHint())

    def translateUI(self):
        self.pagelist.setFixedWidth(self.pagelist.sizeHintForColumn(0) + 12)
        self.setWindowTitle(app.caption(_("Preferences")))

    def trySaveSettings(self):
        try:
            self.saveSettings()
        except CancelClosingPreferences:
            return False
        else:
            return True

    def maybeAccept(self):
        if (not self.buttons.button(QDialogButtonBox.StandardButton.Apply).isEnabled()
              or self.trySaveSettings()):
            self.accept()

    def done(self, result):
        # save our size and selected page
        global _prefsindex
        _prefsindex = self.pagelist.currentRow()
        super().done(result)

    def pages(self):
        """Yields the settings pages that are already instantiated."""
        for n in range(self.stack.count()):
            yield self.stack.widget(n)

    def showHelp(self):
        userguide.show(self.pagelist.currentItem().help)

    def loadSettings(self):
        """Loads the settings on reset."""
        for page in self.pages():
            page.loadSettings()
            page.hasChanges = False
        self.buttons.button(QDialogButtonBox.StandardButton.Apply).setEnabled(False)

    def saveSettings(self):
        """Saves the settings and applies them."""
        for page in self.pages():
            if page.hasChanges:
                page.saveSettings() # this may raise CancelClosingPreferences
                page.hasChanges = False
        self.buttons.button(QDialogButtonBox.StandardButton.Apply).setEnabled(False)

        # emit the signal
        app.settingsChanged()

    def setCurrentPage(self, id):
        for i in range(self.pagelist.count()):
            item = self.pagelist.item(i)
            if item.help == id:
                self.pagelist.setCurrentRow(i)
                return

    def slotCurrentItemChanged(self, item):
        item.activate()

    def changed(self):
        """Call this to enable the Apply button."""
        self.buttons.button(QDialogButtonBox.StandardButton.Apply).setEnabled(True)


class PrefsItemBase(QListWidgetItem):
    help = "preferences"
    def __init__(self):
        super().__init__()
        self._widget = None
        self.setIcon(icons.get(self.iconName))
        app.translateUI(self)

    def activate(self):
        dlg = self.listWidget().parentWidget()
        if self._widget is None:
            w = self._widget = self.widget(dlg)
            dlg.stack.addWidget(w)
            w.loadSettings()
            w.changed.connect(dlg.changed)
            w.changed.connect(w.markChanged)
        dlg.stack.setCurrentWidget(self._widget)


class General(PrefsItemBase):
    help = "prefs_general"
    iconName = "preferences-system"
    def translateUI(self):
        self.setText(_("General"))

    def widget(self, dlg):
        from . import general
        return general.GeneralPrefs(dlg)


class MusicViewers(PrefsItemBase):
    help = "prefs_musicviewers"
    iconName = "Audio-x-generic"
    def translateUI(self):
        self.setText(_("Music View"))

    def widget(self, dlg):
        from . import musicviewers
        return musicviewers.MusicViewers(dlg)


class LilyPond(PrefsItemBase):
    help = "prefs_lilypond"
    iconName = "lilypond-run"
    def translateUI(self):
        self.setText(_("LilyPond"))

    def widget(self, dlg):
        from . import lilypond
        return lilypond.LilyPondPrefs(dlg)


class Midi(PrefsItemBase):
    help = "prefs_midi"
    iconName = "audio-volume-medium"
    def translateUI(self):
        self.setText(_("MIDI"))

    def widget(self, dlg):
        from . import midi
        return midi.MidiPrefs(dlg)


class Helpers(PrefsItemBase):
    help = "prefs_helpers"
    iconName = "applications-other"
    def translateUI(self):
        self.setText(_("Helper Apps"))

    def widget(self, dlg):
        from . import helpers
        return helpers.Helpers(dlg)


class Paths(PrefsItemBase):
    help = "prefs_paths"
    iconName = "folder-open"
    def translateUI(self):
        self.setText(_("Paths"))

    def widget(self, dlg):
        from . import paths
        return paths.Paths(dlg)


class Documentation(PrefsItemBase):
    help = "prefs_lilydoc"
    iconName = "help-contents"
    def translateUI(self):
        self.setText(_("LilyPond Docs"))

    def widget(self, dlg):
        from . import documentation
        return documentation.Documentation(dlg)


class Shortcuts(PrefsItemBase):
    help = "prefs_shortcuts"
    iconName = "preferences-desktop-keyboard-shortcuts"
    def translateUI(self):
        self.setText(_("Shortcuts"))

    def widget(self, dlg):
        from . import shortcuts
        return shortcuts.Shortcuts(dlg)


class Editor(PrefsItemBase):
    help = "prefs_editor"
    iconName = "document-properties"
    def translateUI(self):
        self.setText(_("Editor"))

    def widget(self, dlg):
        from . import editor
        return editor.Editor(dlg)


class FontsColors(PrefsItemBase):
    help = "prefs_fontscolors"
    iconName = "applications-graphics"
    def translateUI(self):
        self.setText(_("Fonts & Colors"))

    def widget(self, dlg):
        from . import fontscolors
        return fontscolors.FontsColors(dlg)


class Tools(PrefsItemBase):
    help = "prefs_tools"
    iconName = "preferences-other"
    def translateUI(self):
        self.setText(_("Tools"))

    def widget(self, dlg):
        from . import tools
        return tools.Tools(dlg)


class Extensions(PrefsItemBase):
    help = "prefs_extensions"
    iconName = "network-plug"
    def translateUI(self):
        self.setText(_("Extensions"))

    def widget(self, dlg):
        from . import extensions
        return extensions.Extensions(dlg)


class Page(QWidget):
    """Base class for settings pages."""
    changed = pyqtSignal()
    hasChanges = False

    def markChanged(self):
        """Called when something changes in the dialog."""
        self.hasChanges = True

    def loadSettings(self):
        """Should load settings from config into our widget."""

    def saveSettings(self):
        """Should write settings from our widget to config."""


class ScrolledPage(Page):
    """Base class for settings pages that are scrollable.

    Te scrolledWidget attribute has the widget the other components
    can be added to.

    """
    def __init__(self, dialog):
        super().__init__(dialog)
        layout = QVBoxLayout(contentsMargins=QMargins(0, 0, 0, 0), spacing=0)
        self.setLayout(layout)
        scrollarea = QScrollArea(frameWidth=0, frameShape=QScrollArea.Shape.NoFrame)
        layout.addWidget(scrollarea)
        self.scrolledWidget = QWidget(scrollarea)
        scrollarea.setWidget(self.scrolledWidget)
        scrollarea.setWidgetResizable(True)


class GroupsPage(Page):
    """Base class for a Page with SettingsGroups.

    The load and save methods of the SettingsGroup groups are automatically called.

    """
    def __init__(self, dialog):
        super().__init__(dialog)
        self.groups = []

    def loadSettings(self):
        for group in self.groups:
            group.loadSettings()

    def saveSettings(self):
        for group in self.groups:
            group.saveSettings()

    def group(self, classname):
        """Access another group on the current page
        by passing its class."""
        for g in self.groups:
            if isinstance(g, classname):
                return g
        raise ValueError("No Group with class \"{}\" on this Page")


class ScrolledGroupsPage(GroupsPage, ScrolledPage):
    def __init__(self, dialog):
        ScrolledPage.__init__(self, dialog)
        self.groups = []


class Group(QGroupBox):
    """This is a QGroupBox that auto-adds itself to a Page."""
    changed = pyqtSignal()

    def __init__(self, page):
        super().__init__()
        self._page = page
        page.groups.append(self)
        self.changed.connect(page.changed)

    def loadSettings(self):
        """Should load settings from config into our widget."""

    def saveSettings(self):
        """Should write settings from our widget to config."""

    def page(self):
        """Return a reference to the Page the Group is on."""
        return self._page

    def siblingGroup(self, classname):
        """Return a reference to another Group on the same page."""
        return self.page().group(classname)



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The global things in Frescobaldi.
"""



import os
import sys
import platform
import importlib.util
import weakref

from PyQt6.QtCore import QObject, QSettings, Qt, QThread, QStandardPaths
from PyQt6.QtWidgets import QApplication, QMenuBar

### needed for QWebEngine
### it wants those two things be done before constructing QApplication()
if importlib.util.find_spec('PyQt6.QtWebEngineWidgets'):
    import PyQt6.QtWebEngineWidgets
    QApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts, True)
### end needed for QWebEngine

import appinfo

qApp = None                     # instantiate() puts the QApplication obj. here
windows = []
documents = []

from signals import Signal, SignalContext

# signals
appInstantiated  = Signal()     # Called when the QApplication is instantiated
appStarted = Signal()           # Called when the main event loop is entered
aboutToQuit = Signal()          # Use this and not qApp.aboutToQuit
mainwindowCreated = Signal()    # MainWindow
mainwindowClosed = Signal()     # MainWindow
documentCreated = Signal()      # Document
documentUrlChanged = Signal()   # Document
documentLoaded = Signal()       # Document
documentModificationChanged = Signal() # Document
documentClosed = Signal()       # Document
documentSaved = Signal()        # Document
documentSaving = SignalContext() # Document
viewCreated = Signal()          # View
viewSpaceCreated = Signal()     # ViewSpace (see viewmanager.py)
languageChanged = Signal()      # (no arguments)
settingsChanged = Signal()      # (no arguments)
sessionChanged = Signal()       # (name)
saveSessionData = Signal()      # (name)
jobStarted = Signal()           # (Document, Job)
jobFinished = Signal()          # (Document, Job, bool success)


def activeWindow():
    """Return the currently active MainWindow.

    Only returns None if there are no windows at all.

    """
    if windows:
        w = QApplication.activeWindow()
        if w in windows:
            return w
        return windows[0]

def openUrl(url, encoding=None):
    """Returns a Document instance for the given QUrl.

    If there is already a document with that url, it is returned.

    """
    d = findDocument(url)
    if not d:
        # special case if there is only one document:
        # if that is empty and unedited, use it.
        if (len(documents) == 1
            and documents[0].url().isEmpty()
            and documents[0].isEmpty()
            and not documents[0].isUndoAvailable()
            and not documents[0].isRedoAvailable()):
            d = documents[0]
            d.setEncoding(encoding)
            if not url.isEmpty():
                d.load(url)
        else:
            import document
            d = document.EditorDocument.new_from_url(url, encoding)
    return d

def findDocument(url):
    """Returns a Document instance for the given QUrl if already loaded.

    Returns None if no document with given url exists or if the url is empty.

    """
    if not url.isEmpty():
        for d in documents:
            if url == d.url():
                return d

def instantiate():
    """Instantiate the global QApplication object."""
    global qApp
    args = list(map(os.fsencode, [os.path.abspath(sys.argv[0])] + sys.argv[1:]))
    if platform.system() == "Windows":
        args.append("-platform")
        args.append("windows:fontengine=freetype")
    qApp = QApplication(args)
    QApplication.setApplicationName(appinfo.name)
    QApplication.setApplicationVersion(appinfo.version)
    QApplication.setDesktopFileName(appinfo.desktop_file_name)
    QApplication.setOrganizationName(appinfo.name)
    QApplication.setOrganizationDomain(appinfo.domain)
    if platform.system() == "Darwin":
        qApp._menubar = QMenuBar()
    appInstantiated()

def oninit(func):
    """Call specified function on QApplication instantiation.

    If the QApplication already has been instantiated, the function is called
    directly.

    As this function returns the specified function, you can use this as a
    decorator.

    """
    if qApp:
        func()
    else:
        appInstantiated.connect(func)
    return func

def run():
    """Enter the Qt event loop."""
    result = qApp.exec()
    aboutToQuit()
    return result

def restart():
    """Restarts Frescobaldi."""
    args = [os.path.abspath(sys.argv[0])] + sys.argv[1:]
    python_executable = sys.executable
    if python_executable:
        args = [python_executable] + args
    #NOTE: This deliberately uses subprocess and not job.Job
    # It has turned out that using Job the code files are not
    # reloaded at all.
    import subprocess
    subprocess.Popen(args)

def _make_retranslate_callback(obj):
    ref = weakref.ref(obj)
    def retranslate():
        found_obj = ref()
        if found_obj is None:
            # obj was already collected. Now remove this retranslator itself,
            # we don't need it anymore.
            languageChanged.disconnect(retranslate)
            return
        found_obj.translateUI()
    return retranslate

def translateUI(obj, priority=0):
    """Translates texts in the object.

    Texts are translated again if the language is changed.
    The object must have a translateUI() method.  It is
    also called by this function.

    This function only causes weak references to the object to be ingested, so
    it does not prevent garbage-collecting the object, and will just not
    retranslate it after the object is garbage-collected.  Furthermore, if the
    object is a QObject, the function will refrain from retranslating it as soon
    as the underlying C++ object is deleted, even if the Python object itself is
    only garbage-collected later.  This means that a widget can safely
    retranslate its subwidgets in its translateUI() method, since they are still
    alive on the C++ level, as the widget itself is alive.  On the other hand,
    for non-QObjects, keep in mind the contract: the object must be able to
    retranslate itself without errors at any time until it is garbage-collected.

    """
    obj.translateUI()
    retranslator = _make_retranslate_callback(obj)
    languageChanged.connect(retranslator, priority)
    if isinstance(obj, QObject):
        obj.destroyed.connect(lambda: languageChanged.disconnect(retranslator))

def caption(title):
    """Returns a nice dialog or window title with appname appended."""
    return f"{title} \u2013 {appinfo.appname}"

def filetypes(extension=None):
    """Returns a list of supported filetypes.

    If a type matches extension, it is placed first.

    """
    have, havenot = [], []
    for patterns, name in (
            ("{0} (*.ly *.lyi *.ily)",          _("LilyPond Files")),
            ("{0} (*.tex *.lytex *.latex)",     _("LaTeX Files")),
            ("{0} (*.docbook *.lyxml)",         _("DocBook Files")),
            ("{0} (*.html *.xml)",              _("HTML Files")),
            ("{0} (*.itely *.tely *.texi *.texinfo)", _("Texinfo Files")),
            ("{0} (*.scm)",                     _("Scheme Files")),
            ("{0} (*)",                         _("All Files")),
            ):
        if extension and extension in patterns:
            have.append(patterns.format(name))
        else:
            havenot.append(patterns.format(name))
    return ";;".join(have + havenot)

def basedir():
    """Returns a base directory for documents.

    First looks in the session settings, then the default settings.
    If no directory was set, this returns the user's Documents directory.

    """
    import sessions
    conf = sessions.currentSessionGroup()
    if not conf:
        conf = QSettings()
    basedir = conf.value("basedir", "", str)
    return basedir if basedir else QStandardPaths.writableLocation(
        QStandardPaths.StandardLocation.DocumentsLocation)

def settings(name):
    """Returns a QSettings object referring a file in ~/.config/frescobaldi/"""
    s = QSettings()
    s.beginGroup(name)
    s.setFallbacksEnabled(False)
    return s

def excepthook(exctype, excvalue, exctb):
    """Called when a Python exception goes unhandled."""
    from traceback import format_exception
    sys.stderr.write(''.join(format_exception(exctype, excvalue, exctb)))
    if exctype is not KeyboardInterrupt:
        # show dialog, but not when in non-GUI thread
        if QThread.currentThread() == qApp.thread():
            import exception
            exception.ExceptionDialog(exctype, excvalue, exctb)

def displayhook(obj):
    """Prevent normal displayhook from overwriting __builtin__._"""
    if obj is not None:
        print(repr(obj))


_extensions = None

def load_extensions(mainwindow):
    """Called from MainWindow.__init__()"""
    import extensions
    global _extensions
    _extensions = extensions.Extensions(mainwindow)

def extensions():
    """Return the global Extensions object."""
    global _extensions
    return _extensions


_job_queue = None

def job_queue():
    """Return the global JobQueue object."""
    global _job_queue
    if _job_queue is None:
        import job.queue
        #TODO: save the number of runners in the Preferences
        #NOTE: Provide code for *changing* the number of runners
        _job_queue = job.queue.GlobalJobQueue()
    return _job_queue

_is_git_controlled = None

def is_git_controlled():
    global _is_git_controlled
    if _is_git_controlled is None:
        import vcs
        _is_git_controlled = vcs.app_is_git_controlled()
    return _is_git_controlled

_editor_font = None

def editor_font(requested_family=None):
    """Returns a font suitable for editing text as a QFont object.

    If a specific font family is requested and available, it will be used.
    Otherwise, this attempts to pick a readable monospace font (the exact
    choice of which is platform-dependent).

    """
    global _editor_font
    if requested_family or _editor_font is None:
        from PyQt6.QtGui import QFontDatabase
        available_families = QFontDatabase.families()
    if _editor_font is None:
        # This is always a safe choice but not necessarily the best available
        _editor_font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        # Our preferred font is generally the default of the platform's native
        # text editor (Notepad, TextEdit, etc.). If this has changed over time
        # then we try the most recent first.
        if platform.system() == "Darwin":   # macOS
            preferred_families = ["SF Mono", "Menlo", "Monaco"]
        elif platform.system() == "Windows":
            # Avoid Lucida Console since it lacks a bold weight
            preferred_families = ["Consolas"]
        else:
            preferred_families = ["monospace"]
        for family in preferred_families:
            if family in available_families:
                _editor_font.setFamily(family)
                break
    # Caller requests should not override the global default
    font = _editor_font
    if requested_family and requested_family in available_families:
        font.setFamily(requested_family)
    return font



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The Score Wizard dialog.
"""


from PyQt6.QtCore import pyqtSignal, QSettings, QUrl
from PyQt6.QtWidgets import (
    QDialog, QDialogButtonBox, QGroupBox, QTabWidget, QVBoxLayout, QWidget)

import app
import indent
import qutil
import userguide
import ly.document
import ly.dom
import ly.music
import ly.util


class ScoreWizardDialog(QDialog):

    pitchLanguageChanged = pyqtSignal(str)

    def __init__(self, mainwindow):
        super().__init__(mainwindow)
        self.addAction(mainwindow.actionCollection.help_whatsthis)
        self._pitchLanguage = None
        self.mainwindow = mainwindow

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.tabs = QTabWidget()
        b = self.dialogButtons = QDialogButtonBox()
        b.setStandardButtons(
            QDialogButtonBox.StandardButton.Reset
            | QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        b.accepted.connect(self.accept)
        b.rejected.connect(self.reject)
        userguide.addButton(b, "scorewiz")
        b.button(QDialogButtonBox.StandardButton.Reset).clicked.connect(self.reset)
        self.previewButton = b.addButton('', QDialogButtonBox.ButtonRole.ActionRole)
        self.previewButton.clicked.connect(self.showPreview)
        layout.addWidget(self.tabs)
        layout.addWidget(b)

        self.header = Header(self)
        self.tabs.addTab(self.header, '')
        self.parts = Parts(self)
        self.tabs.addTab(self.parts, '')
        self.settings = Settings(self)
        self.tabs.addTab(self.settings, '')

        self.tabs.setCurrentIndex(0)
        self.tabs.widget(0).widget() # activate it
        self.tabs.currentChanged.connect(self.slotCurrentChanged)
        qutil.saveDialogSize(self, "scorewiz/dialog/size")
        app.translateUI(self)
        self.accepted.connect(self.slotAccepted)

    def translateUI(self):
        self.setWindowTitle(app.caption(_("Score Setup Wizard")))
        for i in range(self.tabs.count()):
            self.tabs.setTabText(i, self.tabs.widget(i).title())
        self.dialogButtons.button(QDialogButtonBox.StandardButton.Reset).setText(_("Clear"))
        self.dialogButtons.button(QDialogButtonBox.StandardButton.Reset).setToolTip(_(
            "Clears the current page of the Score Wizard."))
        self.previewButton.setText(_("Preview"))

    def slotCurrentChanged(self, i):
        """Lazy-loads the tab's page if shown for the first time."""
        self.tabs.widget(i).widget()

    def reset(self):
        self.tabs.currentWidget().widget().clear()

    def resetAll(self):
        for tab in self.header, self.parts, self.settings:
            tab.widget().clear()

    def setPitchLanguage(self, language):
        if language != self._pitchLanguage:
            self._pitchLanguage = language
            self.pitchLanguageChanged.emit(language)

    def pitchLanguage(self):
        if self._pitchLanguage is None:
            # load setting; saving occurs in .settings.py
            lang = QSettings().value('scorewiz/lilypond/pitch_language', '', str)
            from .scoreproperties import keyNames
            if lang not in keyNames:
                lang = ''
            self._pitchLanguage = lang
        return self._pitchLanguage

    def slotAccepted(self):
        """Makes the score and puts it in the editor."""
        from . import build
        builder = build.Builder(self)       # get the builder
        doc = builder.document()            # get the ly.dom document tree
        if not self.settings.widget().generalPreferences.relpitch.isChecked():
            # remove pitches from \relative commands
            for n in doc.find(ly.dom.Relative):
                for n1 in n.find(ly.dom.Pitch, 1):
                    n.remove(n1)
        text = builder.text(doc)            # convert to LilyPond source text
        lydoc = ly.document.Document(text)  # temporarily store it in a lydoc
        cursor = ly.document.Cursor(lydoc)  # make a cursor selecting it
        indent.indenter().indent(cursor)    # indent it according to user prefs
        doc = app.openUrl(QUrl())           # get a new Frescobaldi document
        doc.setPlainText(lydoc.plaintext()) # write the text in it
        doc.setModified(False)              # make it "not modified"
        self.parent().setCurrentDocument(doc)

    def showPreview(self):
        """Shows a preview."""
        # get the document and fill in some example music
        from . import preview, build
        builder = build.Builder(self)
        doc = builder.document()
        preview.examplify(doc)
        # preview it
        import musicpreview
        dlg = musicpreview.MusicPreviewDialog(self)
        dlg.preview(builder.text(doc), _("Score Preview"))
        dlg.exec()
        dlg.cleanup()

    def readScore(self):
        """Read the score of an existing document."""
        self.resetAll()
        cursor = self.mainwindow.textCursor()
        text = cursor.document().toPlainText()
        # Parse the music
        music = ly.music.document(ly.document.Document(text))
        for item in music:
            if isinstance(item, ly.music.items.Header):
                self.header.readFromMusicItem(item)
            elif isinstance(item, ly.music.items.Assignment):
                name = item.name()
                if name == 'global':
                    self.settings.readFromMusicItem(item)
                elif name.endswith('Part'):
                    self.parts.readFromMusicItem(item)


class Page(QWidget):
    """A Page in the tab widget.

    Basically this is just a QWidget that loads the desired page
    as soon as the widget() is called for the first time.

    """
    def __init__(self, dialog):
        super().__init__(dialog)
        self._widget = None

    def title(self):
        """Should return a title."""

    def widget(self):
        if self._widget is None:
            layout = QVBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            self.setLayout(layout)
            w = self._widget = self.createWidget(self)
            layout.addWidget(w)
        return self._widget

    def createWidget(self, parent):
        """Should return the widget for this tab."""


class Header(Page):
    def title(self):
        return _("&Titles and Headers")

    def createWidget(self, parent):
        from . import header
        return header.HeaderWidget(parent)

    def readFromMusicItem(self, headerBlock):
        """Read an existing header block as a ly.music.items.Header."""
        widget = self.widget()
        for item in headerBlock:
            name = item.name()
            value = item.value()
            if name in widget.edits:
                widget.edits[name].setText(value.plaintext())


class Parts(Page):
    def __init__(self, dialog):
        super().__init__(dialog)
        # Directory mapping assignment identifiers to the corresponding
        # Part class. For example: self._partTypes['jazzGuitar'] == JazzGuitar
        self._partTypes = {}

    def title(self):
        return _("&Parts")

    def createWidget(self, parent):
        from . import score
        return score.ScorePartsWidget(parent)

    def readFromMusicItem(self, assignment):
        """Read a part definition from an ly.music.items.Assignment."""
        from . import score
        widget = self.widget()
        if not self._partTypes:
            # This is only needed when reading from an existing score,
            # so generate it the first time it is used.
            from . import build, parts
            for category in parts.categories:
                for item in category.items:
                    self._partTypes[ly.util.mkid(item.__name__)] = item
        name = assignment.name()
        # Make sure this is, in fact, a part definition before proceeding.
        # We already check this in ScoreWizardDialog.readScore(), but it never
        # hurts to be safe...
        if name.endswith('Part'):
            try:
                # TODO: Handle containers
                parent = widget.scoreView
                part = self._partTypes[name[:-4]]
                box = QGroupBox(widget.partSettings)
                score.PartItem(parent, part, box)
            except KeyError:
                pass


class Settings(Page):
    def title(self):
        return _("&Score settings")

    def createWidget(self, parent):
        from . import settings
        return settings.SettingsWidget(parent)

    def readFromMusicItem(self, assignment):
        """Read settings from an ly.music.items.Assignment."""
        from . import scoreproperties
        widget = self.widget()
        sp = widget.scoreProperties
        for item in assignment.value():
            if isinstance(item, ly.music.items.KeySignature):
                pitch = item.pitch()
                # pitch.alter is a fractions.Fraction
                key = (pitch.note, pitch.alter.numerator)
                if key in scoreproperties.keys:
                    sp.keyNote.setCurrentIndex(scoreproperties.keys.index(key))
                for mode, translation in scoreproperties.modes:
                    if mode == item.mode():
                        sp.keyMode.setCurrentText(translation())
                        break
            elif isinstance(item, ly.music.items.Partial):
                length = item.partial_length()
                midiDuration = (length.denominator, length.numerator)
                if midiDuration in scoreproperties.midiDurations:
                    # index 0 is "None"
                    sp.pickup.setCurrentIndex(
                        scoreproperties.midiDurations.index(midiDuration) + 1
                    )
            elif isinstance(item, ly.music.items.Tempo):
                fraction = item.fraction()
                midiDuration = (fraction.denominator, fraction.numerator)
                if midiDuration in scoreproperties.midiDurations:
                    sp.metronomeNote.setCurrentIndex(
                        scoreproperties.midiDurations.index(midiDuration)
                    )
                tempo = item.tempo()
                if tempo:
                    sp.metronomeValue.setCurrentText(str(tempo[0]))
                if item.text():
                    sp.tempo.setText(item.text().plaintext())
            elif isinstance(item, ly.music.items.TimeSignature):
                # Note item.fraction().numerator is always 1
                fraction = "{}/{}".format(item.numerator(),
                                            item.fraction().denominator)
                sp.timeSignature.setCurrentText(fraction)





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Keyboard shortcuts settings page.
"""


from PyQt6.QtCore import QSettings
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QRadioButton,
    QStyleFactory,
    QTabWidget,
    QVBoxLayout,
    QWidget
)

import app
import appinfo
import icons
import preferences
import sessions
import util
import i18n
import remote
import language_names

from widgets.urlrequester import UrlRequester


class GeneralPrefs(preferences.ScrolledGroupsPage):
    def __init__(self, dialog):
        super().__init__(dialog)

        layout = QVBoxLayout()
        self.scrolledWidget.setLayout(layout)

        layout.addWidget(General(self))
        layout.addWidget(SessionsAndFiles(self))
        layout.addWidget(ExperimentalFeatures(self))

        # Since this is the first page shown, its size hint determines the
        # initial size of the whole dialog. We want it to be big enough to
        # show all the items on this page without scrolling
        self._sizeHint = layout.minimumSize()
        # Pad the width so wider pages like Tools won't scroll horizontally
        self._sizeHint.setWidth(self._sizeHint.width() + 40)

    def sizeHint(self):
        return self._sizeHint


class General(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        grid = QGridLayout()
        self.setLayout(grid)

        self.langLabel = QLabel()
        self.lang = QComboBox(currentIndexChanged=self.changed)
        grid.addWidget(self.langLabel, 0, 0)
        grid.addWidget(self.lang, 0, 1)

        self.styleLabel = QLabel()
        self.styleCombo = QComboBox(currentIndexChanged=self.changed)
        grid.addWidget(self.styleLabel, 1, 0)
        grid.addWidget(self.styleCombo, 1, 1)

        self.systemIcons = QCheckBox(toggled=self.changed)
        grid.addWidget(self.systemIcons, 2, 0, 1, 3)
        self.tabsClosable = QCheckBox(toggled=self.changed)
        grid.addWidget(self.tabsClosable, 3, 0, 1, 3)
        self.splashScreen = QCheckBox(toggled=self.changed)
        grid.addWidget(self.splashScreen, 4, 0, 1, 3)
        self.allowRemote = QCheckBox(toggled=self.changed)
        grid.addWidget(self.allowRemote, 5, 0, 1, 3)

        grid.setColumnStretch(2, 1)

        # fill in the language combo
        self._langs = ["C", ""]
        self.lang.addItems(('', ''))
        langnames = [(language_names.languageName(lang, lang), lang) for lang in i18n.available()]
        langnames.sort()
        for name, lang in langnames:
            self._langs.append(lang)
            self.lang.addItem(name)

        # fill in style combo
        self.styleCombo.addItem('')
        self.styleCombo.addItems(QStyleFactory.keys())

        app.translateUI(self)

    def loadSettings(self):
        s = QSettings()
        lang = s.value("language", "", str)
        try:
            index = self._langs.index(lang)
        except ValueError:
            index = 1
        self.lang.setCurrentIndex(index)
        style = s.value("guistyle", "", str).lower()
        styles = [name.lower() for name in QStyleFactory.keys()]
        try:
            index = styles.index(style) + 1
        except ValueError:
            index = 0
        self.styleCombo.setCurrentIndex(index)
        self.systemIcons.setChecked(s.value("system_icons", True, bool))
        self.tabsClosable.setChecked(s.value("tabs_closable", True, bool))
        self.splashScreen.setChecked(s.value("splash_screen", True, bool))
        self.allowRemote.setChecked(remote.enabled())

    def saveSettings(self):
        s = QSettings()
        s.setValue("language", self._langs[self.lang.currentIndex()])
        s.setValue("system_icons", self.systemIcons.isChecked())
        s.setValue("tabs_closable", self.tabsClosable.isChecked())
        s.setValue("splash_screen", self.splashScreen.isChecked())
        s.setValue("allow_remote", self.allowRemote.isChecked())
        if self.styleCombo.currentIndex() == 0:
            s.remove("guistyle")
        else:
            s.setValue("guistyle", self.styleCombo.currentText())
        # update all top-level windows, so icon changes are picked up
        for w in QApplication.topLevelWidgets():
            if w.isVisible():
                w.update()

    def translateUI(self):
        self.setTitle(_("Language and Style"))
        self.langLabel.setText(_("Language:"))
        self.lang.setItemText(0, _("No Translation"))
        self.lang.setItemText(1, _("System Default Language (if available)"))
        self.styleLabel.setText(_("Style:"))
        self.styleCombo.setItemText(0, _("Default"))
        self.systemIcons.setText(_("Use System Icons"))
        self.systemIcons.setToolTip(_(
            "If checked, icons of the desktop icon theme "
            "will be used instead of the bundled icons."))
        self.splashScreen.setText(_("Show Splash Screen on Startup"))
        self.tabsClosable.setText(_("Show Close Button on Document tabs"))
        self.allowRemote.setText(_("Open Files in Running Instance"))
        self.allowRemote.setToolTip(_(
            "If checked, files will be opened in a running Frescobaldi "
            "application if available, instead of starting a new instance."))


class SessionsAndFiles(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        def changed():
            self.changed.emit()
            self.new_combo.setEnabled(self.template.isChecked())
            self.session_combo.setEnabled(self.session_custom.isChecked())

        def customchanged():
            self.changed.emit()
            self.filenameTemplate.setEnabled(self.customFilename.isChecked())

        self.verbose_toolbuttons = QCheckBox(toggled=self.changed)
        layout.addWidget(self.verbose_toolbuttons)

        self.tabs = QTabWidget()
        layout.addWidget(self.tabs)

        # New Documents Tab
        self.new_tab = QWidget()
        self.tabs.addTab(self.new_tab, "")
        new_layout_wrap = QVBoxLayout()
        self.new_tab.setLayout(new_layout_wrap)
        new_layout = QGridLayout()
        new_layout_wrap.addLayout(new_layout)

        self.emptyDocument = QRadioButton(toggled=changed)
        self.lilyVersion = QRadioButton(toggled=changed)
        self.template = QRadioButton(toggled=changed)
        self.new_combo = QComboBox(currentIndexChanged=changed)

        new_layout.addWidget(self.emptyDocument, 0, 0, 1, 2)
        new_layout.addWidget(self.lilyVersion, 1, 0, 1, 2)
        new_layout.addWidget(self.template, 2, 0, 1, 1)
        new_layout.addWidget(self.new_combo, 2, 1, 1, 1)
        new_layout_wrap.addStretch()

        # Saving Files Tab
        self.save_tab = QWidget()
        self.tabs.addTab(self.save_tab, "")
        save_layout = QVBoxLayout()
        self.save_tab.setLayout(save_layout)

        self.stripwsp = QCheckBox(toggled=self.changed)
        self.backup = QCheckBox(toggled=self.changed)
        self.metainfo = QCheckBox(toggled=self.changed)
        self.format = QCheckBox(toggled=self.changed)
        save_layout.addWidget(self.stripwsp)
        save_layout.addWidget(self.backup)
        save_layout.addWidget(self.metainfo)
        save_layout.addWidget(self.format)
        basedir_layout = QHBoxLayout()
        save_layout.addLayout(basedir_layout)

        self.basedirLabel = QLabel()
        self.basedir = UrlRequester()
        basedir_layout.addWidget(self.basedirLabel)
        basedir_layout.addWidget(self.basedir)
        self.basedir.changed.connect(self.changed)

        filename_layout = QHBoxLayout()
        save_layout.addLayout(filename_layout)

        self.customFilename = QCheckBox(toggled=customchanged)
        self.filenameTemplate = QLineEdit(textEdited=self.changed)
        filename_layout.addWidget(self.customFilename)
        filename_layout.addWidget(self.filenameTemplate)

        # Sessions Tab
        self.session_tab = QWidget()
        self.tabs.addTab(self.session_tab, "")
        session_layout = QGridLayout()
        session_layout_wrap = QVBoxLayout()
        self.session_tab.setLayout(session_layout_wrap)

        self.session_label = QLabel()
        session_layout_wrap.addWidget(self.session_label)
        session_layout_wrap.addLayout(session_layout)

        self.session_none = QRadioButton(toggled=changed)
        self.session_lastused = QRadioButton(toggled=changed)
        self.session_custom = QRadioButton(toggled=changed)
        self.session_combo = QComboBox(currentIndexChanged=changed)

        session_layout.addWidget(self.session_none, 0, 0, 1, 2)
        session_layout.addWidget(self.session_lastused, 1, 0, 1, 2)
        session_layout.addWidget(self.session_custom, 2, 0, 1, 1)
        session_layout.addWidget(self.session_combo, 2, 1, 1, 1)
        session_layout_wrap.addStretch()

        self.loadNewCombo()
        self.page().parent().finished.connect(self.saveTabIndex)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Sessions and Files"))
        self.verbose_toolbuttons.setText(_(
            "Add pull-down menus in main toolbar"
        ))
        self.verbose_toolbuttons.setToolTip("<font>{}</font>".format(
            _("If set, the file related buttons in the main toolbar will "
            "provide pull-down menus with additional functions.")
        ))

        # New Documents Tab
        self.tabs.setTabText(0, _("New Document"))
        self.emptyDocument.setText(_("Create an empty document"))
        self.lilyVersion.setText(_("Create a document that contains the LilyPond version statement"))
        self.template.setText(_("Create a document from a template:"))
        from snippet import snippets
        for i, name in enumerate(self._names):
            self.new_combo.setItemText(i, snippets.title(name))

        # Saving Files Tab
        self.tabs.setTabText(1, _("Saving"))
        self.stripwsp.setText(_("Strip trailing whitespace"))
        self.stripwsp.setToolTip(_(
            "If checked, Frescobaldi will remove unnecessary whitespace at the "
            "end of lines (but not inside multi-line strings)."))
        self.backup.setText(_("Keep backup copy"))
        self.backup.setToolTip(_(
            "Frescobaldi always backups a file before overwriting it "
            "with a new version.\n"
            "If checked those backup copies are retained."))
        self.metainfo.setText(_("Remember cursor position, bookmarks, etc."))
        self.format.setText(_("Format document"))

        self.basedirLabel.setText(_("Default directory:"))
        self.basedirLabel.setToolTip(_("The default folder for your LilyPond documents (optional)."))
        self.customFilename.setText(_("Use custom default file name:"))
        self.customFilename.setToolTip(_(
            "If checked, Frescobaldi will use the template to generate a default file name.\n"
            "{title} and {composer} will be replaced by title and composer of that document."))

        # Sessions Tab
        self.tabs.setTabText(2, _("Sessions"))
        self.session_label.setText(_(
            "Session to load if Frescobaldi is started without arguments"
        ))
        self.session_none.setText(_("Start with no session"))
        self.session_lastused.setText(_("Start with last used session"))
        self.session_custom.setText(_("Start with session:"))

    def loadNewCombo(self):
        from snippet import snippets
        self._names = [name for name in snippets.names()
                        if snippets.get(name).variables.get('template')]
        self.new_combo.clear()
        self.new_combo.addItems([''] * len(self._names))

    def loadSettings(self):
        s = QSettings()
        self.verbose_toolbuttons.setChecked(
            s.value("verbose_toolbuttons", False, bool)
        )

        # New Documents Tab
        ndoc = s.value("new_document", "empty", str)
        template = s.value("new_document_template", "", str)
        if template in self._names:
            self.new_combo.setCurrentIndex(self._names.index(template))
        if ndoc == "template":
            self.template.setChecked(True)
        elif ndoc == "version":
            self.lilyVersion.setChecked(True)
        else:
            self.emptyDocument.setChecked(True)

        # Saving Files Tab
        self.stripwsp.setChecked(s.value("strip_trailing_whitespace", False, bool))
        self.backup.setChecked(s.value("backup_keep", False, bool))
        self.metainfo.setChecked(s.value("metainfo", True, bool))
        self.format.setChecked(s.value("format", False, bool))
        self.basedir.setPath(s.value("basedir", "", str))
        self.customFilename.setChecked(s.value("custom_default_filename", False, bool))
        self.filenameTemplate.setText(s.value("default_filename_template", "{composer}-{title}", str))
        self.filenameTemplate.setEnabled(self.customFilename.isChecked())

        # Sessions Tab
        s.beginGroup("session")
        startup = s.value("startup", "none", str)
        if startup ==  "lastused":
            self.session_lastused.setChecked(True)
        elif startup == "custom":
            self.session_custom.setChecked(True)
        else:
            self.session_none.setChecked(True)
        sessionNames = sessions.sessionNames()
        self.session_combo.clear()
        self.session_combo.addItems(sessionNames)
        custom = s.value("custom", "", str)
        if custom in sessionNames:
            self.session_combo.setCurrentIndex(sessionNames.index(custom))
        s.endGroup()
        self.tabs.setCurrentIndex(
            s.value("prefs_general_file_tab_index", 0, int)
        )

    def saveSettings(self):
        s = QSettings()
        s.setValue("verbose_toolbuttons", self.verbose_toolbuttons.isChecked())

        # New Documents Tab
        if self._names and self.template.isChecked():
            s.setValue("new_document", "template")
            s.setValue("new_document_template", self._names[self.new_combo.currentIndex()])
        elif self.lilyVersion.isChecked():
            s.setValue("new_document", "version")
        else:
            s.setValue("new_document", "empty")

        # Saving Files Tab
        s.setValue("strip_trailing_whitespace", self.stripwsp.isChecked())
        s.setValue("backup_keep", self.backup.isChecked())
        s.setValue("metainfo", self.metainfo.isChecked())
        s.setValue("format", self.format.isChecked())
        s.setValue("basedir", self.basedir.path())
        s.setValue("custom_default_filename", self.customFilename.isChecked())
        s.setValue("default_filename_template", self.filenameTemplate.text())

        # Sessions Tab
        s.beginGroup("session")
        s.setValue("custom", self.session_combo.currentText())
        if self.session_custom.isChecked():
            startup = "custom"
        elif self.session_lastused.isChecked():
            startup = "lastused"
        else:
            startup = "none"
        s.setValue("startup", startup)

    def saveTabIndex(self):
        s = app.settings("")
        s.setValue("prefs_general_file_tab_index", self.tabs.currentIndex())


class ExperimentalFeatures(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.experimentalFeatures = QCheckBox(toggled=self.changed)
        layout.addWidget(self.experimentalFeatures)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Experimental Features"))
        self.experimentalFeatures.setText(_("Enable Experimental Features"))
        self.experimentalFeatures.setToolTip('<qt>' + _(
            "If checked, features that are not yet finished are enabled.\n"
            "You need to restart Frescobaldi to see the changes."))

    def loadSettings(self):
        s = QSettings()
        self.experimentalFeatures.setChecked(s.value("experimental-features", False, bool))

    def saveSettings(self):
        s = QSettings()
        s.setValue(
            "experimental-features", self.experimentalFeatures.isChecked()
        )



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Function for helping the user to report bugs.
"""

from PyQt6.QtCore import QUrl, QUrlQuery

import helpers
import appinfo
import debuginfo

def new_github_issue(title, body):
    """
    Open a web brower on a page to create a new issue on GitHub.

    Information about the versions of Frescobaldi and its dependencies will
    be appended to the body.

    The body will be sent in the query part of a URL. It is therefore advisable
    not to make it too long. As of this writing, GitHub imposes a limit around
    6000 characters.
    """
    body += "\n\n" + debuginfo.version_info_string()
    url = QUrl(appinfo.issues_url)
    query = QUrlQuery()
    query.addQueryItem("title", title)
    query.addQueryItem("body", body)
    url.setQuery(query)
    helpers.openUrl(url)

def email(subject, body, recipient=None):
    """Opens the e-mail composer with the given subject and body, with version information added to it."""
    subject = f"[{appinfo.appname} {appinfo.version}] {subject}"
    body = "{}\n\n{}\n\n".format(debuginfo.version_info_string('\n'), body)
    address = recipient or appinfo.maintainer_email
    url = QUrl("mailto:" + address)
    query = QUrlQuery()
    query.addQueryItem("subject", subject)
    query.addQueryItem("body", body)
    url.setQuery(query)
    helpers.openUrl(url, "email")



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Per-tool preferences.
"""


import re

from PyQt6.QtCore import QSettings, Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import (
    QAbstractItemView, QCheckBox, QDoubleSpinBox, QFontComboBox, QHBoxLayout,
    QLabel, QPushButton, QVBoxLayout, QWidget)

import app
import userguide
import qutil
import preferences
import widgets.dialog
import widgets.listedit
import documentstructure


class Tools(preferences.ScrolledGroupsPage):
    def __init__(self, dialog):
        super().__init__(dialog)

        layout = QVBoxLayout()
        self.scrolledWidget.setLayout(layout)

        layout.addWidget(LogTool(self))
        layout.addWidget(CharMap(self))
        layout.addWidget(DocumentList(self))
        layout.addWidget(Outline(self))
        layout.addStretch(1)


class LogTool(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.fontLabel = QLabel()
        self.fontChooser = QFontComboBox(currentFontChanged=self.changed)
        self.fontSize = QDoubleSpinBox(valueChanged=self.changed)
        self.fontSize.setRange(6.0, 32.0)
        self.fontSize.setSingleStep(0.5)
        self.fontSize.setDecimals(1)

        box = QHBoxLayout()
        box.addWidget(self.fontLabel)
        box.addWidget(self.fontChooser, 1)
        box.addWidget(self.fontSize)
        layout.addLayout(box)

        self.showlog = QCheckBox(toggled=self.changed)
        layout.addWidget(self.showlog)

        self.rawview = QCheckBox(toggled=self.changed)
        layout.addWidget(self.rawview)

        self.hideauto = QCheckBox(toggled=self.changed)
        layout.addWidget(self.hideauto)

        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("LilyPond Log"))
        self.fontLabel.setText(_("Font:"))
        self.showlog.setText(_("Show log when a job is started"))
        self.rawview.setText(_("Display plain log output"))
        self.rawview.setToolTip(_(
            "If checked, Frescobaldi will not shorten filenames in the log output."))
        self.hideauto.setText(_("Hide automatic engraving jobs"))
        self.hideauto.setToolTip(_(
            "If checked, Frescobaldi will not show the log for automatically\n"
            "started engraving jobs (LilyPond->Auto-engrave)."))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("log")
        font = QFont(s.value("fontfamily", "monospace", str))
        font.setPointSizeF(s.value("fontsize", 9.0, float))
        with qutil.signalsBlocked(self.fontChooser, self.fontSize):
            self.fontChooser.setCurrentFont(font)
            self.fontSize.setValue(font.pointSizeF())
        self.showlog.setChecked(s.value("show_on_start", True, bool))
        self.rawview.setChecked(s.value("rawview", True, bool))
        self.hideauto.setChecked(s.value("hide_auto_engrave", False, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("log")
        s.setValue("fontfamily", self.fontChooser.currentFont().family())
        s.setValue("fontsize", self.fontSize.value())
        s.setValue("show_on_start", self.showlog.isChecked())
        s.setValue("rawview", self.rawview.isChecked())
        s.setValue("hide_auto_engrave", self.hideauto.isChecked())


class CharMap(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)

        self.fontLabel = QLabel()
        self.fontChooser = QFontComboBox(currentFontChanged=self.changed)
        self.fontSize = QDoubleSpinBox(valueChanged=self.changed)
        self.fontSize.setRange(6.0, 32.0)
        self.fontSize.setSingleStep(0.5)
        self.fontSize.setDecimals(1)

        box = QHBoxLayout()
        box.addWidget(self.fontLabel)
        box.addWidget(self.fontChooser, 1)
        box.addWidget(self.fontSize)
        layout.addLayout(box)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Special Characters"))
        self.fontLabel.setText(_("Font:"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("charmaptool")
        font = self.font()
        family = s.value("fontfamily", "", str)
        if family:
            font.setFamily(family)
        font.setPointSizeF(s.value("fontsize", font.pointSizeF(), float))
        with qutil.signalsBlocked(self.fontChooser, self.fontSize):
            self.fontChooser.setCurrentFont(font)
            self.fontSize.setValue(font.pointSizeF())

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("charmaptool")
        s.setValue("fontfamily", self.fontChooser.currentFont().family())
        s.setValue("fontsize", self.fontSize.value())


class DocumentList(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)
        self.groupCheck = QCheckBox(toggled=self.changed)
        layout.addWidget(self.groupCheck)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Documents"))
        self.groupCheck.setText(_("Group documents by directory"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("document_list")
        self.groupCheck.setChecked(s.value("group_by_folder", False, bool))

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("document_list")
        s.setValue("group_by_folder", self.groupCheck.isChecked())


class Outline(preferences.Group):
    def __init__(self, page):
        super().__init__(page)

        layout = QVBoxLayout()
        self.setLayout(layout)
        self.label = QLabel()
        self.patternList = OutlinePatterns()
        self.patternList.listBox.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self.defaultButton = QPushButton(clicked=self.reloadDefaults)
        self.patternList.layout().addWidget(self.defaultButton, 3, 1)
        self.patternList.layout().addWidget(self.patternList.listBox, 0, 0, 5, 1)
        self.patternList.changed.connect(self.changed)
        self.labelComments = QLabel()
        self.patternListComments = OutlinePatterns()
        self.patternListComments.listBox.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self.defaultButtonComments = QPushButton(clicked=self.reloadDefaultsComments)
        self.patternListComments.layout().addWidget(self.defaultButtonComments, 3, 1)
        self.patternListComments.layout().addWidget(self.patternListComments.listBox, 0, 0, 5, 1)
        self.patternListComments.changed.connect(self.changed)
        layout.addWidget(self.label)
        layout.addWidget(self.patternList)
        layout.addWidget(self.labelComments)
        layout.addWidget(self.patternListComments)
        app.translateUI(self)

    def translateUI(self):
        self.setTitle(_("Outline"))
        self.defaultButton.setText(_("Default"))
        self.defaultButton.setToolTip(_("Restores the built-in outline patterns."))
        self.label.setText(_("Patterns to match in text (excluding comments) that are shown in outline:"))
        self.defaultButtonComments.setText(_("Default"))
        self.defaultButtonComments.setToolTip(_("Restores the built-in outline patterns."))
        self.labelComments.setText(_("Patterns to match in text (including comments) that are shown in outline:"))

    def reloadDefaults(self):
        self.patternList.setValue(documentstructure.default_outline_patterns)

    def reloadDefaultsComments(self):
        self.patternListComments.setValue(documentstructure.default_outline_patterns_comments)

    def loadSettings(self):
        s = QSettings()
        s.beginGroup("documentstructure")
        try:
            patterns = s.value("outline_patterns", documentstructure.default_outline_patterns, str)
        except TypeError:
            patterns = []
        try:
            patterns_comments = s.value("outline_patterns_comments", documentstructure.default_outline_patterns_comments, str)
        except TypeError:
            patterns_comments = []
        self.patternList.setValue(patterns)
        self.patternListComments.setValue(patterns_comments)

    def saveSettings(self):
        s = QSettings()
        s.beginGroup("documentstructure")
        if self.patternList.value() != documentstructure.default_outline_patterns:
            s.setValue("outline_patterns", self.patternList.value())
        else:
            s.remove("outline_patterns")
        if self.patternListComments.value() != documentstructure.default_outline_patterns_comments:
            s.setValue("outline_patterns_comments", self.patternListComments.value())
        else:
            s.remove("outline_patterns_comments")


class OutlinePatterns(widgets.listedit.ListEdit):
    def openEditor(self, item):
        dlg = widgets.dialog.TextDialog(self,
            _("Enter a regular expression to match:"),
            app.caption("Outline"))
        userguide.addButton(dlg.buttonBox(), "outline_configure")
        dlg.setValidateFunction(is_regex)
        dlg.setText(item.text())
        if dlg.exec():
            item.setText(dlg.text())
            return True
        return False


def is_regex(text):
    """Return True if text is a valid regular expression."""
    try:
        re.compile(text, re.M)
    except re.error:
        return False
    return True



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Provides version information of important supporting modules.
"""


import functools
import platform
import os

import app
import appinfo


def _catch_unknown(f):
    """Decorate a function, returning "unknown" on import/attribute error."""
    @functools.wraps(f)
    def wrapper():
        try:
            return f()
        except (ImportError, AttributeError):
            return "unknown"
    return wrapper


@_catch_unknown
def app_version():
    import appinfo
    return appinfo.version

@_catch_unknown
def pyqt_version():
    import PyQt6.QtCore
    return PyQt6.QtCore.PYQT_VERSION_STR

@_catch_unknown
def qt_version():
    import PyQt6.QtCore
    return PyQt6.QtCore.QT_VERSION_STR

@_catch_unknown
def python_version():
    return platform.python_version()

@_catch_unknown
def operating_system():
    plat = platform.platform()
    if platform.system() == "Linux":
        try:
            distro = platform.freedesktop_os_release()["PRETTY_NAME"]
        except OSError:
            # play it safe
            distro = "unknown distribution"
        return f"{plat} ({distro})"
    else:
        return plat

@_catch_unknown
def ly_version():
    import ly.pkginfo
    return ly.pkginfo.version

@_catch_unknown
def qpageview_version():
    import qpageview
    return qpageview.version_string

if platform.system() == "Darwin":
    @_catch_unknown
    def mac_installation_kind():
        import macos
        if macos.inside_lightweight_app_bundle():
            return 'lightweight .app bundle'
        elif macos.inside_app_bundle():
            return 'standalone .app bundle'
        else:
            return 'command line'

elif platform.system() == "Linux":
    @_catch_unknown
    def linux_installation_kind():
        import linux
        if linux.inside_flatpak():
            return "Flatpak"
        else:
            return "distro package or installed from source"

def version_info_named():
    """Yield all the relevant names and their version string."""
    yield appinfo.appname, appinfo.version
    yield "Extension API", appinfo.extension_api
    yield "Python", python_version()
    if app.is_git_controlled():
        import vcs
        repo = vcs.app_repo
        yield "Git branch", repo.active_branch()
        commit = repo.run_command(
            'log',
            ['-n', '1', '--format=format:%h'])
        yield "on commit", commit[0]
    yield "python-ly", ly_version()
    yield "Qt", qt_version()
    yield "PyQt", pyqt_version()
    yield "qpageview", qpageview_version()
    yield "OS", operating_system()
    if platform.system() == 'Darwin':
        yield "installation kind", mac_installation_kind()
    elif platform.system() == "Linux":
        yield "Installation kind", linux_installation_kind()


def version_info_string(separator='\n'):
    """Return all version names as a string, joint with separator."""
    return separator.join(map("{0[0]}: {0[1]}".format, version_info_named()))



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.


"""
Show a dialog with available text and music fonts.
"""

from PyQt6.QtCore import (
    QByteArray,
    QSettings,
    Qt
)
from PyQt6.QtWidgets import (
    QApplication,
    QDialog,
    QDialogButtonBox,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

import app
import fonts
import userguide
import widgets.dialog

from . import (
    fontcommand,
    musicfonts,
    preview,
    textfonts
)


_default_fonts = {
    'music': 'emmentaler',
    'brace': 'emmentaler',
    'roman': 'TeXGyre Schola',
    'sans': 'TeXGyre Heros',
    'typewriter': 'TeXGyre Cursor'
}


class FontsDialog(QDialog):
    """Dialog to show available fonts"""

    _selected_fonts = {
        'music': _default_fonts['music'],
        'brace': _default_fonts['brace'],
        'roman': _default_fonts['roman'],
        'sans': _default_fonts['sans'],
        'typewriter': _default_fonts['typewriter']
    }

    def __init__(
        self,
        parent,
        lilypond_info=None,
        show_music=True
    ):
        QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
        super().__init__(
            parent,
        )
        self.info = lilypond_info
        if not lilypond_info:
            import documentinfo
            self.info = documentinfo.lilyinfo(parent.currentDocument())

        self.result = ''
        self.available_fonts = fonts.available(self.info)

        # Notation fonts (and preview) are limited to LilyPond >= 2.19.12
        # At some point we may remove the old dialog altogether
        # and instead make this dialog behave differently
        # (i.e. hide the music font stuff and use old font selection code)
        # self.show_music = self.info.version() >= (2, 19, 12).
        # For now this distinction is made by the action and simply
        # the dialog to be used is chosen. At some point the old
        # "Set document fonts" dialog should be dropped.
        #
        # Also, it may at some point be indicated to make this
        # dialog usable to *only* choose text fonts, e.g. from
        # the "Fonts & Colors" Preference page.
        #
        # NOTE: All facilities that *seemed* to support this functionality
        # have been removed to avoid confusion.
        self.show_music = show_music

        # Basic dialog attributes
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.setWindowModality(Qt.WindowModality.WindowModal)

        layout = QVBoxLayout()
        self.setLayout(layout)

        # Create a QSplitter as main widget
        self.splitter = QSplitter()
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        layout.addWidget(self.splitter)

        # Left side layout
        self.left_column = QWidget()
        left_layout = QVBoxLayout()
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_layout.setSpacing(0)
        self.left_column.setLayout(left_layout)
        self.splitter.addWidget(self.left_column)

        # Status area
        # TODO: Populate that widget with a QStackedLayout where
        # different widgets are shown corresponding to the visible tab.
        self.status_area = QWidget()
        left_layout.addWidget(self.status_area)

        # Create the QTabWidget for the dialog's left part
        self.tab_widget = QTabWidget()
        left_layout.addWidget(self.tab_widget)

        # Text Fonts Tab
        self.tab_widget.addTab(textfonts.TextFontsWidget(self), '')
        # Music Fonts Tab
        self.tab_widget.addTab(musicfonts.MusicFontsWidget(self), '')
        # Show/configure the generated font setting command
        self.font_command_tab = fontcommand.FontCommandWidget(self)
        self.tab_widget.addTab(self.font_command_tab, '')
        # Show various fontconfig information
        self.tab_widget.addTab(
            textfonts.MiscFontsInfoWidget(self.available_fonts), '')

        # Create the RHS score preview pane.
        self.preview_pane = preview.FontsPreviewWidget(self)
        self.splitter.addWidget(self.preview_pane)

        # Bottom area: button box
        self._button_box = bb = QDialogButtonBox()
        layout.addWidget(bb)
        self.restore_button = bb.addButton(QDialogButtonBox.StandardButton.RestoreDefaults)
        self.copy_button = bb.addButton(QDialogButtonBox.StandardButton.Save)
        self.insert_button = bb.addButton(QDialogButtonBox.StandardButton.Ok)
        self.close_button = bb.addButton(QDialogButtonBox.StandardButton.Close)
        # Add and connect help button
        userguide.addButton(self._button_box, "documentfonts")

        app.translateUI(self)
        self.loadSettings()
        self.connectSignals()

        # Trigger the generation of a preview
        self.invalidate_command()

    def connectSignals(self):
        self.finished.connect(self.saveSettings)
        self.restore_button.clicked.connect(self.restore)
        self.copy_button.clicked.connect(self.copy_result)
        self.insert_button.clicked.connect(self.insert_result)
        self.close_button.clicked.connect(self.close)

    def translateUI(self):
        self.setWindowTitle(app.caption(_("Document Fonts")))
        self.copy_button.setText(_("&Copy"))
        self.copy_button.setToolTip(_("Copy font command to clipboard"))
        self.insert_button.setText(_("&Use"))
        self.insert_button.setToolTip(
            _("Insert font command at the current cursor position")
        )
        self.tab_widget.setTabText(0, _("Text Fonts"))
        self.tab_widget.setTabText(1, _("Music Fonts"))
        self.tab_widget.setTabText(2, _("Font Command"))
        self.tab_widget.setTabText(3, _("Miscellaneous"))

    def loadSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')

        self.select_font('music', s.value('music-font', 'emmentaler', str))
        self.select_font('brace', s.value('brace-font', 'emmentaler', str))
        self.select_font('roman', s.value('roman-font', 'TeXGyre Schola', str))
        self.select_font('sans', s.value('sans-font', 'TeXGyre Heros', str))
        self.select_font(
            'typewriter', s.value('typewriter-font', 'TeXGyre Cursor', str)
        )
        self.splitter.restoreState(
            s.value('splitter-state', QByteArray(), QByteArray)
        )

    def select_font(self, family, name):
        self._selected_fonts[family] = name

    def selected_font(self, family):
        return self._selected_fonts[family]

    def saveSettings(self):
        s = QSettings()
        s.beginGroup('document-fonts-dialog')

        s.setValue('music-font', self.selected_font('music'))
        s.setValue('brace-font', self.selected_font('brace'))
        s.setValue('roman-font', self.selected_font('roman'))
        s.setValue('sans-font', self.selected_font('sans'))
        s.setValue('typewriter-font', self.selected_font('typewriter'))

        # Dialog layout
        s.setValue('splitter-state', self.splitter.saveState())

    def copy_result(self):
        """Copies the font command (as shown) to the clipboard."""
        from PyQt6.QtGui import QGuiApplication
        cmd = self.font_cmd()
        if cmd[-1] != '\n':
            cmd = cmd + '\n'
        QGuiApplication.clipboard().setText(cmd)
        self.accept()

    def font_cmd(self, approach=None):
        """Return the font setting command as shown in the Font Command tab."""
        approach = approach or self.font_command_tab.approach
        return self.font_command_tab.command(approach)

    def font_full_cmd(self, approach=None):
        """Return the "full" command with all properties/fonts."""
        approach = approach or self.font_command_tab.approach
        return self.font_command_tab.full_cmd(approach)

    def insert_result(self):
        """Inserts the font command (as shown) at the current position"""
        self.result = self.font_cmd()
        self.accept()

    def invalidate_command(self):
        """Triggers a regeneration of the font command and new display of the
        example (if necessary).
        """
        self.font_command_tab.invalidate_command()

    def restore(self):
        """Reset fonts to defaults"""
        for name in _default_fonts:
            self.select_font(name, _default_fonts[name])
        self.invalidate_command()

    def show_sample(self):
        """Call the preview pane to show the current sample."""
        self.preview_pane.show_sample()



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The extensions framework
"""

import importlib
import os
import pathlib
import sys
from time import perf_counter
from traceback import extract_tb

from PyQt6.QtCore import (
    QDir,
    QObject,
    QSettings,
    QSize,
    Qt
)
from PyQt6.QtGui import (
    QFont,
    QIcon,
    QStandardItem,
    QStandardItemModel
)
from PyQt6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QDialog,
    QHeaderView,
    QMenu,
    QStyle,
    QTreeView
)

import app
import icons
import vbcl
import qsettings
import util
from . import actions, settings


class Extension(QObject):
    """
    Base class for all Extension objects.
    Ensures integration in the Tools menu and other places,
    and provides common APIs both for the extension to
    access resources and for the application to interact
    with the extension.

    An Extension provides a number of convenience functions to
    objects that could also be retrieved through other means,
    but we want to provide a convenient API for extension development.
    They can be found at the end of this implementation.
    """

    # To be defined in subclasses,
    # used for the Tools menu submenu title and (later)
    # in the Preferences page.
    _display_name = "Display name of extension not specified"

    # To be defined in subclasses,
    # specify a class for the extension's action collection.
    # If no action collection class is given the default
    # action collection will be empty, which *may* be
    # desired if an extension only wants to provide a Tool panel.
    _action_collection_class = actions.ExtensionActionCollection

    # Can be defined in subclasses to create a Tool panel.
    # If a class is provided for a panel widget a Tool panel is
    # automatically created. By default the initial dock area is
    # the left dock area, but this can also be specified by
    # overriding the corresponding class variable.
    _panel_widget_class = None
    _panel_dock_area = Qt.DockWidgetArea.LeftDockWidgetArea

    # Can be defined in subclasses to provide a configuration widget.
    # If this is present it will be shown in the Preferences page.
    # Config widgets can be any QWidget descendants, but must fulfill
    # some requirements:
    # - They have to  implement the load_settings() and save_settings() methods.
    # - Instead of a QSettings() object they have to interact with
    #   a self.settings() object that is injected from this class.
    # - Any changes that have to be "applied" (i.e. that should enable
    #   the "Apply" button) must not be saved immediately but emit the
    #   self.parent().changed signal to ensure consistent behaviour
    #   in the Preferences dialog.
    _config_widget_class = None

    # Dictionary of known settings has to be defined in subclasses.
    # Keys are strings - the names of settings.
    # Values are default values to be used, also defining the type, e.g.:
    #   { 'name': 'NN', 'max-entries': 999 }
    _settings_config = {}

    def __init__(self, parent, name):
        """Initialize (load) the extension.
        NOTE: This loading takes place at program start, so
        any concrete extension should take care to load its
        resources (dialogs etc.) only upon use. During __init__
        only the actions and the optional Tool panel should be
        created.
        The parent argument points to the global Extensions object."""
        assert issubclass(
            self._action_collection_class, actions.ExtensionActionCollection)
        super().__init__(parent)
        self._name = name
        self._root_directory = os.path.join(
            parent.root_directory(), self._name)
        self._settings = settings.ExtensionSettings(self)
        self._action_collection =  self._action_collection_class(self)
        self._menus = {}
        self._panel = None
        self._config_widget = None
        self._icon = None
        self._metadata = None
        self._load_time = ''
        self.create_panel()

        # Hook that can be implemented to update extension status
        # after settings change.
        app.settingsChanged.connect(self.app_settings_changed)

    def action_collection(self):
        """Return the extension's action collection."""
        return self._action_collection

    def app_settings_changed(self):
        """Implement to respond to global settings changes
        (e.g. to update GUI or action status).
        This should only respond to changed settings *outside*
        the extension (done in the Preferences dialog). Settings
        changes of the extension itself invoke
        Extension.settings_changed(key, old, new) instead."""
        pass

    def config_widget(self, preference_group):
        """Return a config widget if provided by the extension, or None.
        Do *not* use caching (this is done in the Preferences group),
        set the preference group as the widget's parent,
        inject extension() and settings() attributes."""
        if not self._config_widget_class:
            return None
        widget = self._config_widget_class(preference_group)
        widget.extension = lambda: self
        widget.settings = lambda: self.settings()
        return widget

    def create_panel(self):
        """Create the Extension's Tool Panel if a widget class is provided."""
        if self._panel_widget_class:
            from . import panel
            self._panel = panel.ExtensionPanel(
                self,
                self._panel_widget_class,
                self._panel_dock_area)

    def display_name(self):
        """Return the display name of the extension, or the extension
        key if metadata has not been loaded properly."""
        return self.infos().get('extension-name', self.name())

    def has_icon(self):
        """Returns True if the extension has a custom icon."""
        return self.icon() is not None

    def icon(self):
        """Return the extension's Icon, or None."""
        return self.parent().icon(self.name())

    def infos(self):
        """Return the infos dictionary for the extension,
        or an empty dictionary if they could not be loaded."""
        return self.parent().infos(self.name()) or {}

    def load_time(self):
        """Return the loading time for the Extension object, formatted
        as a string with milliseconds."""
        return self._load_time

    def menu(self, target):
        """Return a submenu for use in an Extensions menu,
        or None if there should be no entry for the given target.
        target may be 'tools' for the main Tools=>Extensions menu,
        or a known keyword for one of the context menus (see
        actions.ExtensionActionCollection.__init__ for the available keys).

        If the target is 'tools' and the extension exports a Panel
        then this will be inserted at the top of the menu, otherwise
        only the action list will be used to create the menu.

        actions.ExtensionActionCollection.set_menu_action_list() is used
        to define the various menu structures that can include 'flat'
        actions and hierarchical submenus.
        """

        def add_action_or_submenu(m, entry):
            if isinstance(entry, QMenu):
                m.addMenu(entry)
            else:
                m.addAction(entry)

        m = self._menus.get(target, None)
        if not m:
            # Menu is not cached already, create it
            actions = self.action_collection().menu_actions(target)
            if target != 'tools' and not actions:
                # empty context menu
                return None
            panel = self.panel()
            # An extension without at least one action or panel is invalid
            if not (panel or actions):
                m = QMenu(_("Invalid extension menu: {name}").format(
                    name=self.display_name()))
                return m

            # finally create the menu
            m = self._menus[target] = QMenu(
                self.display_name(), self.mainwindow())
            if self.has_icon():
                m.setIcon(self.icon())
            if panel and target == 'tools':
                panel.toggleViewAction().setText(_("&Tool Panel"))
                m.addAction(panel.toggleViewAction())
                if actions:
                    m.addSeparator()
            for entry in actions:
                add_action_or_submenu(m, entry)
        return m

    def name(self):
        """Return the extension's internal name
        (which actually is retrieved from the directory name)."""
        return self._name

    def panel(self):
        """Return the extension's Tool panel, if available, None otherwise."""
        return self._panel

    def root_directory(self):
        """Return the extension's root directory."""
        return self._root_directory

    def set_load_time(self, ms):
        """Store the time needed to load the application.
        A string with two fractional digits in 'ms'."""
        self._load_time = ms

    def settings(self):
        """Reference to the extension's settings object."""
        return self._settings

    def settings_changed(self, key, old, new):
        """Called when a setting *within the extension* has been changed.
        Complements app_settings_changed.
        Can be implemented in subclasses."""
        pass

    def widget(self):
        """Returns the extension's Tool Panel Widget, if available. or None."""
        return self.panel().widget() if self.panel() else None

    # Convenience functions to access elements of the application
    # and the current documents. This will be extended.

    def current_document(self):
        """Returns the currently opened document or None."""
        return self.mainwindow().currentDocument()

    def mainwindow(self):
        """Returns a reference to the main window."""
        return self.parent().mainwindow()

    def text_cursor(self):
        """The text cursor of the current document."""
        return self.mainwindow().textCursor()


class ExtensionMixin(QObject):
    """Convenience mixin class for widgets in an extension.

    The class provides additional attributes to directly interact with the
    elements of an extension, such as the main Extension object, the settings,
    the mainwindow etc.

    If the 'parent' passed to `__init__(parent)` has an 'extension()' attribute
    (which is for example true when the widget is instantiated as an
    extension's panel widget) this is implicitly made available within the
    widget.
    Otherwise the used class must specify the '_extension_name' class variable,
    exactly using the name matching the extension directory. This is true for
    additional widgets that may be used *within* the panel widget or in newly
    created dialogs etc.
    """

    _extension_name = ''

    def extension(self):
        """Return the actual Extension object if possible."""
        if self._extension_name:
            return app.extensions().get(self._extension_name)
        elif hasattr(self.parent(), 'extension'):
            # The parent has access to the extension (typically the Panel)
            return self.parent().extension()
        else:
            raise AttributeError(_(
                "Class '{classname}' can't access Extension object. "
                "It should provide an _extension_name class variable."
                ).format(classname=self.__class__.__name__))

# More properties are accessed through the extension() property

    def action_collection(self):
        """Return the extension's action collection."""
        return self.extension().action_collection()

    def current_document(self):
        """Return the current document."""
        return self.extension().current_document()

    def mainwindow(self):
        """Return the active MainWindow."""
        return self.extension().mainwindow()

    def panel(self):
        """Return the extension's panel, or None if the widget
        is not a Panel widget and the extension does not have a panel."""
        return self.extension().panel()

    def settings(self):
        """Return the extension's local settings object."""
        return self.extension().settings()


class FailedModel(QStandardItemModel):
    """Data model that is created to report about failed extensions.
    An extension may fail due to errors in the metadata (extension is
    loaded anyway), unresolvable dependencies, or exceptions while
    instantiating the Extension object.

    This data model is created and populated upon program start if
    loading the extension reveals any problems. It is then stored as
    the Extensions object's _failed_model member and reused to populate
    various FailedTree widgets.
    """

    def __init__(self, extensions, data):
        super().__init__()
        self.setColumnCount(2)
        self.extensions = extensions
        # Store reference to parent item of failed extensions
        self.exceptions_item = None
        self.populate(data)

    def populate(self, data):
        """Populate the data model using the data passed
        from the extensions object.

        The data model has up to three root-level items:
        - Invalid metadata
        - Failed to load
        - Failed dependencies
        """

        # font for use in various Items
        bold = QFont()
        bold.setWeight(QFont.Weight.Bold)

        root = self.invisibleRootItem()
        infos = data['infos']
        if infos:
            # Handle extensions with metadata errors
            infos_item = QStandardItem()
            infos_item.setFont(bold)
            infos_item.setText(_("Invalid metadata:"))
            infos_tooltip = (
                _("Extensions whose extension.cnf file has errors.\n"
                  "They will be loaded nevertheless."))
            infos_item.setToolTip(infos_tooltip)

            root.appendRow(infos_item)
            for info in infos:
                name_item = QStandardItem(info)
                name_item.setToolTip(infos_tooltip)
                icon = self.extensions.icon(info)
                if icon:
                    name_item.setIcon(icon)
                details_item = QStandardItem(infos[info])
                details_item.setToolTip(infos_tooltip)
                infos_item.appendRow([name_item, details_item])

        exceptions = data['exceptions']
        if exceptions:
            # Handle extensions that failed to load properly
            import traceback
            exceptions_item = self.exceptions_item = QStandardItem()
            exceptions_item.setFont(bold)
            exceptions_item.setText(_("Failed to load:"))
            extensions_tooltip = (
                _("Extensions that failed to load properly.\n"
                  "Double click on name to show the stacktrace.\n"
                  "Please contact the extension maintainer."))
            exceptions_item.setToolTip(extensions_tooltip)

            root.appendRow(exceptions_item)
            for ext in exceptions:
                extension_info = self.extensions.infos(ext)
                name = (extension_info.get('extension-name', ext)
                    if extension_info
                    else ext)
                name_item = QStandardItem(name)
                name_item.setToolTip(extensions_tooltip)
                icon = self.extensions.icon(ext)
                if icon:
                    name_item.setIcon(icon)
                exc_info = exceptions[ext]
                # store exception information in the first item
                name_item.exception_info = exc_info
                message = f'{exc_info[0].__name__}: {exc_info[1]}'
                details_item = QStandardItem(message)
                details_item.setToolTip(extensions_tooltip)
                exceptions_item.appendRow([name_item, details_item])

        dependencies = data['dependencies']
        if dependencies:
            # Handle extensions with dependency issues
            dep_item = QStandardItem(_("Failed dependencies:"))
            dep_item.setFont(bold)
            dep_tooltip = (
                _("Extensions with failed or circular dependencies.\n"
                  "They are not loaded."))
            dep_item.setToolTip(dep_tooltip)

            root.appendRow(dep_item)
            missing = dependencies.get('missing', None)
            if missing:
                missing_item = QStandardItem(_("Missing:"))
                missing_item.setFont(bold)
                missing_item.setToolTip(dep_tooltip)
                dep_item.appendRow(missing_item)
                for m in missing:
                    item = QStandardItem(m)
                    item.setToolTip(dep_item)
                    missing_item.appendRow(item)
            inactive = dependencies.get('inactive', None)
            if inactive:
                inactive_item = QStandardItem(_("Inactive:"))
                inactive_item.setFont(bold)
                inactive_item.setToolTip(dep_tooltip)
                dep_item.appendRow(inactive_item)
                for i in inactive:
                    item = QStandardItem(i)
                    item.setToolTip(dep_tooltip)
                    inactive_item.appendRow(item)
            circular = dependencies.get('circular', None)
            if circular:
                circular_item = QStandardItem(_("Circular:"))
                circular_item.setFont(bold)
                circular_item.setToolTip(dep_tooltip)
                dep_item.appendRow(circular_item)
                item = QStandardItem(' | '.join(circular))
                item.setToolTip(dep_tooltip)
                circular_item.appendRow(item)


class FailedTree(QTreeView):
    """QTreeView descendant to display information about failed extensions
    in the initial message dialog and the Preferences panel.
    The strategy is to create the QTreeView object everytime it is used
    without having to re-populate the data model."""

    # QStandardItemModel is set when any failed extensions are detected
    # while loading extensions. So checking for this class variable can
    # determine whether any extensions failed
    _model = None

    def __init__(self, parent=None):
        super().__init__(parent)
        #TODO: Set the height of the QTreeView to 4-6 rows
        #https://stackoverflow.com/questions/53591159/set-the-height-of-a-qtreeview-to-height-of-n-rows
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.setHeaderHidden(True)
        self.setModel(self._model)
        self.header().setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
        self.doubleClicked.connect(self.tree_double_clicked)

    def tree_double_clicked(self, index):
        """Show the details of a failed extension in a message box."""
        model = self.model()
        item = model.itemFromIndex(index)
        parent = item.parent()
        if item and item.parent() == model.exceptions_item:

            # TODO: Change this to the Exception handling message box
            # (with option to write an email to maintainer)

            import traceback
            from PyQt6.QtWidgets import QMessageBox
            import appinfo
            row = item.row()
            name_item = parent.child(row, 0)
            message_item = parent.child(row, 1)
            msg_box = QMessageBox()
            msg_box.setWindowTitle(_(appinfo.appname))
            msg_box.setIcon(QMessageBox.Icon.Information)
            extension_name = name_item.text()
            exception_info = name_item.exception_info
            msg_box.setText("{}\n\n{}".format(
                _("Extension '{name}' failed to load with the given explanation:").format(
                    name=extension_name),
                message_item.text()))
            msg_box.setInformativeText(
                ' '.join(traceback.format_exception(*exception_info)))
            msg_box.exec()


class Extensions(QObject):
    """Global object managing the extensions.
    Accessed with app.extensions()"""

    def __init__(self, mainwindow):
        super().__init__()
        self._mainwindow = mainwindow.objectName()
        # Resources are handled on the global Extensions level because
        # in a number of cases they are independent from an actually
        # loaded Extension object (i.e. have to be functional before loading
        # or with failed extensions too)
        self._menus = {}
        self._icons = {}
        self._infos = {}
        self._extensions = {}
        self._extensions_ordered = []
        self._failed_infos = {}
        self._failed_dependencies = []
        self._failed_extensions = {}

        # These will be set in load_settings():
        self._root_directory = None
        self._active = False
        self._inactive_extensions = []
        self.load_settings()
        app.settingsChanged.connect(self.settings_changed)

        # Only process extensions if root directory is configured properly
        if not (self.root_directory()
            and os.path.isdir(self.root_directory())):
            return
        # Info about *installed* extensions is loaded
        # regardless of 'active' state
        self.load_infos()
        if self.active():
            self._extensions_ordered = self.check_dependencies()
            #if not self._failed_dependencies:
            self.load_extensions()
            if (self._failed_infos
                or self._failed_extensions
                or self._failed_dependencies):
                FailedTree._model = FailedModel(self, {
                    'infos': self._failed_infos,
                    'dependencies': self._failed_dependencies,
                    'exceptions': self._failed_extensions
                })
                self.report_failed()
        del self._failed_infos
        del self._failed_dependencies
        del self._failed_extensions

    def active(self):
        """Returns True if extensions are enabled globally."""
        return self._active

    def check_dependencies(self):
        """Check extension metadata for unresolved or circular dependencies.

        Returns a list of extension keys in topological order.
        Inactive extensions are skipped.
        Extensions with missing dependencies are logged and skipped.
        Circular dependencies are looged and skipped.
        """

        inactive = self.inactive_extensions()
        inactive_dependencies = []
        missing_dependencies = []
        infos = self._infos
        nodes = {}
        incoming = {}
        no_incoming = {}
        result = []

        # Create topological order (and check for cyclic dependencies)
        # by Kahn's algorithm
        # https://en.wikipedia.org/wiki/Topological_sorting#Algorithms
        for ext in self._infos.keys():
            nodes[ext] = { 'incoming': [], 'outgoing': [] }
        # read incoming dependencies,
        # check for missing or deactivated dependencies.
        for ext in self._infos.keys():
            info = infos[ext]
            if not info:
                continue
            deps = info['dependencies']
            if deps == '---':
                continue
            for dep in deps:
                # missing and inactive dependencies are beside Kahn's algorithm
                if dep in self._infos.keys(): #active + inactive:
                    nodes[dep]['incoming'].append(ext)
                    nodes[ext]['outgoing'].append(dep)
                    if dep in inactive:
                        inactive_dependencies.append((ext, dep))
                else:
                    missing_dependencies.append((ext, dep))

        # Create initial set/queue of nodes with no incoming edges
        for n in nodes:
            if nodes[n]['incoming']:
                incoming[n] = nodes[n]
            else:
                no_incoming[n] = nodes[n]

        # Process queue until empty
        while no_incoming:
            ext = list(no_incoming.keys())[0]
            node = no_incoming.pop(ext)
            result.append(ext)
            # Process nodes ext depends on
            for outgoing in node['outgoing']:
                target_node = incoming[outgoing]
                # Remove egde
                target_node['incoming'].remove(ext)
                # If target has no more incoming edges, move to queue
                if not target_node['incoming']:
                    no_incoming[outgoing] = incoming.pop(outgoing)

        if missing_dependencies or inactive_dependencies or incoming:
            self._failed_dependencies = {}
            if missing_dependencies:
                missing = self._failed_dependencies['missing'] = []
                for dep in missing_dependencies:
                    ext = dep[0]
                    if ext in result:
                        result.remove(ext)
                    missing.append(f"{ext} => {dep[1]}")
            if inactive_dependencies:
                inactive = self._failed_dependencies['inactive'] = []
                for dep in inactive_dependencies:
                    ext = dep[0]
                    if ext in result:
                        result.remove(ext)
                    inactive.append(f"{ext} => {dep[1]}")
            if incoming:
                # If there remain incoming edges we have a
                # circular dependency (according to Kahn's algorithm)
                self._failed_dependencies['circular'] = [
                    ext for ext in incoming.keys()]
        # List of extensions with all dependencies resolved,
        # and ordered in correct topological order
        return reversed(result)

    def load_extensions(self):
        """Load active extensions in topological order."""
        root = self.root_directory()
        if root not in sys.path:
            sys.path.append(root)

        for ext in [ext for ext  in self._extensions_ordered
            if ext not in self.inactive_extensions()]:
            try:
                # measure loading time
                start = perf_counter()
                # Try importing the module. Will fail here if there's
                # no Python module in the subdirectory or loading the module
                # produces errors
                module = importlib.import_module(ext)
                # Add extension's icons dir (if present) to icon search path
                icon_path = os.path.join(self.root_directory(), ext, 'icons')
                if os.path.isdir(icon_path):
                    search_paths = QDir.searchPaths('icons')
                    QDir.setSearchPaths('icons', [icon_path] + search_paths)
                # Instantiate the extension,
                # this will fail if the module is no valid extension
                # (doesn't have an Extension class) or has other errors in it
                extension = module.Extension(self, ext)
                end = perf_counter()
                extension.set_load_time(
                    f"{(end - start) * 1000:.2f} ms")
                self._extensions[ext] = extension
            except Exception:
                self._failed_extensions[ext] = sys.exc_info()

    def _load_icon(self, name):
        """Tries to load a main icon for the given extension if
        <extension-dir>/icons/extension.svg exists.
        """
        icon_file_name = os.path.join(
            self.root_directory(), name, 'icons', 'extension.svg')
        self._icons[name] = (
            QIcon(icon_file_name) if os.path.exists(icon_file_name) else None)

    def icon(self, name):
        """Return a main icon for the given extension, or None."""
        icon = self._icons.get(name, False)
        if icon is False: # None would be a valid result indicating 'no icon'
            self._load_icon(name)
        return self._icons[name]

    def _load_infos(self, extension_name):
        """Load an extension's metadata from the extension.cnf file.
        Optional keys are assigned 'empty' default values, while
        missing mandatory keys will cause the load to fail. Such extensions
        will be reported as "Failed metadata", but the extension is
        loaded anyway."""

        # Mandatory keys in the extension.cfg file
        mandatory_config_keys = [
            'extension-name',
            'maintainers',
            'version',
            'api-version',
            'license'
        ]
        # Default values for accepted config keys
        config_defaults = {
            'short-description': '',
            'description': '',
            'repository': '',
            'website': '',
            'dependencies': '---'
        }
        config_file = os.path.join(
            self.root_directory(), extension_name, 'extension.cnf')
        try:
            return vbcl.parse_file(
                config_file,
                mandatory_keys=mandatory_config_keys,
                defaults=config_defaults)
        except Exception as e:
            self._failed_infos[extension_name] = str(e)
            return None

    def load_infos(self):
        """Load extension metadata from all subdirs that include an
        extension.cnf file."""
        root = self.root_directory()
        subdirs = [dir for dir in os.listdir(root) if os.path.isfile(
            os.path.join(root, dir, 'extension.cnf'))]
        for d in subdirs:
            self._infos[d] = self._load_infos(d)

    def extensions(self):
        """Return a list of all loaded extensions, ordered by their
        display name."""
        result = [self._extensions[ext] for ext in self._extensions.keys()]
        result.sort(key=lambda extension: extension.display_name())
        return result

    def get(self, name):
        """Return an Extension object for an extension with the given name,
        or None if such a module doesn't exist."""
        return self._extensions.get(name, None)

    def inactive_extensions(self):
        """Return a list with names of extensions marked as inactive."""
        return self._inactive_extensions

    def infos(self, name):
        return self._infos.get(name, None)

    def installed_extensions(self):
        """Return a list of all installed extensions, independently of their
        load status."""
        result = list(self._infos.keys())
        return sorted(result)

    def is_extension_exception(self, traceback):
        """Check if the given traceback points to an exception occurring
        within an extension. If so, return a tuple with
        - extension name (display name)
        - first maintainer
        - extension key.
        If the exception is *not* from an extension return None"""
        for frame_summary in extract_tb(traceback):
            file_path = pathlib.Path(frame_summary.filename)
            root_path = pathlib.Path(self.root_directory())
            if util.path_is_relative_to(file_path, root_path):
                rel_file_path = file_path.relative_to(root_path)
                extension_name = rel_file_path.parts[0]
                infos = self.infos(extension_name)
                if infos:
                    return (infos['extension-name'],
                            infos['maintainers'][0],
                            extension_name)

    def load_settings(self):
        """Load application-wide settings relating to extensions."""
        s = QSettings()
        s.beginGroup('extension-settings')
        self._active = s.value('active', True, bool)
        self._inactive_extensions = qsettings.get_string_list(s, 'installed/inactive')
        self._root_directory = s.value('root-directory', '', str)

    def mainwindow(self):
        """Reference to the main window."""
        for w in app.windows:
            if w.objectName() == self._mainwindow:
                return w
        return app.windows[0]

    def menu(self, target):
        """Create and return a nested Extensions menu,
        used for the Tools menu or context menus."""

        m = self._menus.get(target, None)
        if not m:
            #TODO: Create mnemonics for the menu entries
            m = self._menus[target] = QMenu(_("&Extensions"), self.mainwindow())
            m.setIcon(icons.get('network-plug'))
            for ext in self.extensions():
                ext_menu = ext.menu(target)
                if ext_menu:
                    m.addMenu(ext_menu)
        return m

    def report_failed(self):
        """Show a warning message box if extension(s) could either not
        be loaded or produced errors while parsing the extension infos."""

        from PyQt6.QtCore import QCoreApplication
        import appinfo
        import qutil
        import widgets.dialog

        dlg = widgets.dialog.Dialog(
            self.mainwindow(),
            icon='warning',
            buttons=('ok',),
            title=appinfo.appname,
            message=_(
                "There were problems loading the extensions.\n"
                "The following details are also available from the "
                "Preferences dialog."))
        dlg.setMainWidget(FailedTree(self.mainwindow()))
        qutil.saveDialogSize(
            dlg, "extensions/error-dialog/size", QSize(600, 300))
        dlg.setGeometry(QStyle.alignedRect(
            Qt.LayoutDirection.RightToLeft,
            Qt.AlignmentFlag.AlignCenter,
            dlg.size(),
            QApplication.primaryScreen().availableGeometry()))
        dlg.exec()

    def reset_inactive(self, inactive):
        """Save the list of inactive extensions."""
        self._inactive_extensions = list(inactive)

    def root_directory(self):
        """Root directory below which extensions are searched for."""
        return self._root_directory

    def set_inactive(self, extension, state=True):
        """Set one extension to active/inactive"""
        if state:
            if extension not in self._inactive_extensions:
                self._inactive_extensions.append(extension)
        else:
            if extension in self._inactive_extensions:
                self._inactive_extensions.remove(extension)

    def set_tools_menu(self, menu):
        """Store a reference to the global Tools menu."""
        self._tools_menu = menu

    def settings_changed(self):
        """Determine if any changes in global extension settings
        requires a restart."""
        s = QSettings()
        s.beginGroup('extension-settings')
        active = s.value('active', True, bool)
        inactive = qsettings.get_string_list(s, 'installed/inactive')
        root = s.value('root-directory', '', str)
        inactive_changed = False
        if len(inactive) != len(self._inactive_extensions):
            inactive_changed = True
        else:
            for ext in inactive:
                if ext not in self._inactive_extensions:
                    inactive_changed = True
                    break
        self.reset_inactive(inactive)
        if (
            active != self.active()
            or root != self.root_directory()
            or inactive_changed
        ):
            from widgets.restartmessage import suggest_restart
            suggest_restart(_('Extensions have changed'))
